#include "Basis.h"
#include <limits>
#include "CTNode.h"
#include "globalFuncs.h"
#include "Matrix.h"
#include "profile.h"

//---------------------------------------------------------------------------------------------------------------

// Basis class

template <typename Float>
Basis<Float>::Basis(const QUIC::Uint& m, const QUIC::Uint& n, const QUIC::OrthoMode& om, const QUIC::Uint& maxRank, QUIC::MatAllocator<Float> * alloc)
: nCols(n), maxRows(maxRank ? maxRank : iDivUp(m, rowPreallocDenom)), data(NULL), lastRowRemoved(noRow), nRows(0), fullList(), orthoMode(om), allocator(alloc)
{
	QUIC::MatAllocator<Float>::pitched_pointer p = allocator->allocate(n, maxRows);
	data = p.addr;
	pitch = p.pitch;
}

template <typename Float>
Basis<Float>::~Basis()
{
	if (data)
	{
		QUIC::MatAllocator<Float>::pitched_pointer p = QUIC::MatAllocator<Float>::make_pitched_pointer(data, pitch);
		allocator->deallocate(p, nCols, maxRows);
	}
}

template <typename Float>
void Basis<Float>::project(QUIC::Matrix<Float> * A, Float * result, Float * scratch, const QUIC::Uint& scratchPitch)
{
	START_TIMER(PROJECTION, CPU);
	QUIC::Matrix<Float>& Ar = *A;
	const QUIC::Uint& projNRows = A->nRows;
	const QUIC::Uint& projNCols = nRows;
	QUIC::Matrix<Float> scratchMat(projNRows, projNCols, scratch, scratchPitch);
	START_TIMER(PROJECTION_MUL, CPU);
	QUIC::Matrix<Float> basisMat;
	getMatrix(basisMat);
	QUIC::Matrix<Float>::mul(*A, basisMat, scratchMat, false, true);
	STOP_TIMER(PROJECTION_MUL, CPU);
	START_TIMER(PROJECTION_NORM, CPU);
	rowNorms(scratchMat, true, result);
	STOP_TIMER(PROJECTION_NORM, CPU);
	STOP_TIMER(PROJECTION, CPU);
}

template <typename Float>
void Basis<Float>::orthogonalizeRows(iterator newRowIters [], Float * scratch, const QUIC::Uint& scratchPitch)
{
	START_TIMER(ORTHOGONALIZATION, CPU);
	Float * newRows [] = {*newRowIters[0], *newRowIters[1]};
	if (newRows[1] < newRows[0]) std::swap(newRows[1], newRows[0]);
	switch (orthoMode)
	{
		case QUIC::MGS:
		{
			for (QUIC::Int i=0; i<2; ++i)
			{
				iterator iter = newRowIters[i];
				Float * newRow = newRows[i];
				for (++iter; iter != end(); ++iter)
				{
					Float * basisRow = *iter;
					Float dotProd = cpuDot(nCols, basisRow, newRow);
					cpuVectorSum(nCols, -1.0f*dotProd, basisRow, newRow);
				}
				Float norm = cpuNorm(nCols, newRow);
				cpuScalarMul(nCols, 1.0f/norm, newRow);
			}
			break;
		}
		case QUIC::UGS:
		{
			assert(newRows[0] == row(size() - 2) && newRows[1] == row(size() - 1));
			const QUIC::Uint nDotProds = nRows - 2;
			const QUIC::Uint rowSize = pitch/sizeof(Float);
			const QUIC::Uint scratchRowSize = scratchPitch/sizeof(Float);
			cpuMatMul(CPU_TRANS, CPU_NO_TRANS, nDotProds, 2, nCols, 1.0f, data, rowSize, newRows[0], rowSize, 0.0f, scratch, scratchRowSize);
			cpuMatMul(CPU_NO_TRANS, CPU_NO_TRANS, nCols, 2, nDotProds, -1.0f, data, rowSize, scratch, scratchRowSize, 1.0f, newRows[0], rowSize);
			Float norm1 = cpuNorm(nCols, newRows[0]);
			cpuScalarMul(nCols, 1.0f/norm1, newRows[0]);
			Float dotProd = cpuDot(nCols, newRows[0], newRows[1]);
			cpuVectorSum(nCols, -1.0f*dotProd, newRows[0], newRows[1]);
			Float norm2 = cpuNorm(nCols, newRows[1]);
			cpuScalarMul(nCols, 1.0f/norm2, newRows[1]);
			break;
		}
		case QUIC::BLOCK_MGS:
		{
			assert(newRows[0] == row(size() - 2) && newRows[1] == row(size() - 1));
			const QUIC::Uint nDotProds = nRows - 2;
			const QUIC::Uint rowSize = pitch/sizeof(Float);
			const QUIC::Uint scratchRowSize = scratchPitch/sizeof(Float);
			for (QUIC::Uint r=0; r<nDotProds; r+=mgsBlockSize)
			{
				QUIC::Uint nBlock = std::min(mgsBlockSize, nDotProds - r);
				Float * block = row(r);
				cpuMatMul(CPU_TRANS, CPU_NO_TRANS, nBlock, 2, nCols, 1.0f, block, rowSize, newRows[0], rowSize, 0.0f, scratch, scratchRowSize);
				cpuMatMul(CPU_NO_TRANS, CPU_NO_TRANS, nCols, 2, nBlock, -1.0f, block, rowSize, scratch, scratchRowSize, 1.0f, newRows[0], rowSize);
			}
			Float norm1 = cpuNorm(nCols, newRows[0]);
			cpuScalarMul(nCols, 1.0f/norm1, newRows[0]);
			Float dotProd = cpuDot(nCols, newRows[0], newRows[1]);
			cpuVectorSum(nCols, -1.0f*dotProd, newRows[0], newRows[1]);
			Float norm2 = cpuNorm(nCols, newRows[1]);
			cpuScalarMul(nCols, 1.0f/norm2, newRows[1]);
			break;
		}
	}
	STOP_TIMER(ORTHOGONALIZATION, CPU);
}

template <typename Float>
void Basis<Float>::deleteRow(CTNode<Float> * n)
{
	CTNode<Float>& N = *n;
	assert(lastRowRemoved == noRow);
	if (orthoMode == QUIC::MGS)
	{
		const Basis<Float>::iterator& basisVector = N.getBasisVector();
		lastRowRemoved = basisVector.index();
		fullList.erase(basisVector);
	}
	else
	{
		assert(begin().index() == size() - 1);
		assert(*fullList.begin() == size() - 1);
		const iterator& basisVector = N.getBasisVector();
		lastRowRemoved = fullList.front();
		copyRow(begin(), basisVector);
		fullList.front() = basisVector.index();
		fullList.erase(basisVector);
	}
	N.deleteBasisVector();
	--nRows;
}

template <typename Float>
void Basis<Float>::copyRow(const iterator& source, const iterator& dest)
{
	cpuMemcpy(nCols, *source, *dest);
}

template <typename Float>
typename Basis<Float>::iterator Basis<Float>::createNewRow()
{
	if (size() == maxRows) realloc();
	QUIC::Uint newRowId;
	if (lastRowRemoved == noRow)
	{
		newRowId = size();
	}
	else
	{
		newRowId = lastRowRemoved;
		lastRowRemoved = noRow;
	}
	++nRows;
	fullList.push_front(newRowId);
	return begin();
}

template <typename Float>
void Basis<Float>::realloc()
{
	START_TIMER(REALLOCATION, CPU);
	QUIC::MatAllocator<Float>::pitched_pointer oldP = QUIC::MatAllocator<Float>::make_pitched_pointer(data, pitch);
	QUIC::Uint oldH = maxRows;
	maxRows <<= 1uLL;
	QUIC::MatAllocator<Float>::pitched_pointer newP = allocator->allocate(nCols, maxRows);
	data = newP.addr;
	pitch = newP.pitch;
	for (QUIC::Uint i=0; i<nRows; ++i) cpuMemcpy(nCols, arrayRow(oldP.addr, i, oldP.pitch), row(i));
	allocator->deallocate(oldP, nCols, oldH);
	STOP_TIMER(REALLOCATION, CPU);
}

template <typename Float>
QUIC::Uint Basis<Float>::mgsBlockSize = 20;

template <typename Float>
const QUIC::Uint Basis<Float>::noRow = std::numeric_limits<QUIC::Uint>::max();

template <typename Float>
QUIC::MatAllocator<Float> Basis<Float>::cpuAllocator;

//---------------------------------------------------------------------------------------------------------------

// BasisGPU class

template <typename Float>
BasisGPU<Float>::BasisGPU(const QUIC::Uint& m, const QUIC::Uint& n, const QUIC::OrthoMode& om, const QUIC::Uint& maxRank)
: Basis(m, n, om, maxRank, &gpuAllocator)
{}

template <typename Float>
BasisGPU<Float>::~BasisGPU()
{}

template <typename Float>
void BasisGPU<Float>::project(QUIC::Matrix<Float> * A, Float * result, Float * scratch, const QUIC::Uint& scratchPitch)
{
	START_TIMER(PROJECTION, GPU);
	QUIC::Matrix<Float>& Ar = *A;
	const QUIC::Uint& projNRows = A->nRows;
	const QUIC::Uint& projNCols = nRows;
	QUIC::MatrixGPU<Float> scratchMat(projNRows, projNCols, scratch, scratchPitch);
	START_TIMER(PROJECTION_MUL, GPU);
	QUIC::MatrixGPU<Float> basisMat;
	getMatrix(basisMat);
	QUIC::MatrixGPU<Float>::mul(*dynamic_cast<QUIC::MatrixGPU<Float> *>(A), basisMat, scratchMat, false, true);
	STOP_TIMER(PROJECTION_MUL, GPU);
	START_TIMER(PROJECTION_NORM, GPU);
	rowNorms(scratchMat, true, result);
	STOP_TIMER(PROJECTION_NORM, GPU);
	STOP_TIMER(PROJECTION, GPU);
}

template <typename Float>
void BasisGPU<Float>::orthogonalizeRows(Basis<Float>::iterator newRowIters [], Float * scratch, const QUIC::Uint& scratchPitch)
{
	START_TIMER(ORTHOGONALIZATION, GPU);
	Float * newRows [] = {*newRowIters[0], *newRowIters[1]};
	if (newRows[1] < newRows[0]) std::swap(newRows[1], newRows[0]);
	switch (orthoMode)
	{		
		case QUIC::MGS:
		{
			for (QUIC::Int i=0; i<2; ++i)
			{
				iterator iter = newRowIters[i];
				Float * newRow = newRows[i];
				for (++iter; iter != end(); ++iter)
				{
					Float * basisRow = *iter;
					Float dotProd = gpuDot(nCols, basisRow, newRow);
					gpuVectorSum(nCols, -1.0f*dotProd, basisRow, newRow);
				}
				Float norm = gpuNorm(nCols, newRow);
				gpuScalarMul(nCols, 1.0f/norm, newRow);
			}
			break;
		}
		case QUIC::UGS:
		{
			assert(newRows[0] == row(size() - 2) && newRows[1] == row(size() - 1));
			const QUIC::Uint nDotProds = nRows - 2;
			const QUIC::Uint rowSize = pitch/sizeof(Float);
			const QUIC::Uint scratchRowSize = scratchPitch/sizeof(Float);
			gpuMatMul(GPU_TRANS, GPU_NO_TRANS, nDotProds, 2, nCols, 1.0f, data, rowSize, newRows[0], rowSize, 0.0f, scratch, scratchRowSize);
			gpuMatMul(GPU_NO_TRANS, GPU_NO_TRANS, nCols, 2, nDotProds, -1.0f, data, rowSize, scratch, scratchRowSize, 1.0f, newRows[0], rowSize);
			Float norm1 = gpuNorm(nCols, newRows[0]);
			gpuScalarMul(nCols, 1.0f/norm1, newRows[0]);
			Float dotProd = gpuDot(nCols, newRows[0], newRows[1]);
			gpuVectorSum(nCols, -1.0f*dotProd, newRows[0], newRows[1]);
			Float norm2 = gpuNorm(nCols, newRows[1]);
			gpuScalarMul(nCols, 1.0f/norm2, newRows[1]);
			break;
		}
		case QUIC::BLOCK_MGS:
		{
			assert(newRows[0] == row(size() - 2) && newRows[1] == row(size() - 1));
			const QUIC::Uint nDotProds = nRows - 2;
			const QUIC::Uint rowSize = pitch/sizeof(Float);
			const QUIC::Uint scratchRowSize = scratchPitch/sizeof(Float);
			for (QUIC::Uint r=0; r<nDotProds; r+=mgsBlockSize)
			{
				QUIC::Uint nBlock = std::min(mgsBlockSize, nDotProds - r);
				Float * block = row(r);
				gpuMatMul(GPU_TRANS, GPU_NO_TRANS, nBlock, 2, nCols, 1.0f, block, rowSize, newRows[0], rowSize, 0.0f, scratch, scratchRowSize);
				gpuMatMul(GPU_NO_TRANS, GPU_NO_TRANS, nCols, 2, nBlock, -1.0f, block, rowSize, scratch, scratchRowSize, 1.0f, newRows[0], rowSize);
			}
			Float norm1 = gpuNorm(nCols, newRows[0]);
			gpuScalarMul(nCols, 1.0f/norm1, newRows[0]);
			Float dotProd = gpuDot(nCols, newRows[0], newRows[1]);
			gpuVectorSum(nCols, -1.0f*dotProd, newRows[0], newRows[1]);
			Float norm2 = gpuNorm(nCols, newRows[1]);
			gpuScalarMul(nCols, 1.0f/norm2, newRows[1]);
			break;
		}
	}
	STOP_TIMER(ORTHOGONALIZATION, GPU);
}

template <typename Float>
void BasisGPU<Float>::copyRow(const Basis<Float>::iterator& source, const Basis<Float>::iterator& dest)
{
	gpuMemcpy(nCols, *source, *dest);
}

template <typename Float>
void BasisGPU<Float>::realloc()
{
	START_TIMER(REALLOCATION, CPU);
	QUIC::MatAllocator<Float>::pitched_pointer oldP = QUIC::MatAllocator<Float>::make_pitched_pointer(data, pitch);
	QUIC::Uint oldH = maxRows;
	maxRows <<= 1uLL;
	QUIC::MatAllocator<Float>::pitched_pointer newP = allocator->allocate(nCols, maxRows);
	data = newP.addr;
	pitch = newP.pitch;
	QUIC_SAFE_CALL(cudaMemcpy2D(newP.addr, newP.pitch, oldP.addr, oldP.pitch, nCols*sizeof(Float), size(), cudaMemcpyDeviceToDevice));
	allocator->deallocate(oldP, nCols, oldH);
	STOP_TIMER(REALLOCATION, CPU);
}

template <typename Float>
QUIC::MatAllocatorGPU<Float> BasisGPU<Float>::gpuAllocator;

//---------------------------------------------------------------------------------------------------------------

// explicit instantiations

INSTANTIATE_CLASS_TEMPLATE(Basis)