#pragma once

#include <list>
#include <cassert>
#include "QUIC.h"
#include "LSDistribution.h"
#include "MatAllocator.h"

#define ROW_PREALLOC_DENOM 8

template <typename Float>
class Basis : public QUIC::Base
{
public:
	class iterator : public std::list<QUIC::Uint>::iterator, public QUIC::Base
	{
	public:
		iterator(const std::list<QUIC::Uint>::iterator& iter, Basis * b);
		iterator& operator++();
		iterator operator++(int);
		Float * operator*() const;
		QUIC::Uint index() const;
	private:
		Basis * basis;
	};
	Basis(const QUIC::Uint& m, const QUIC::Uint& n, const QUIC::OrthoMode& om, const QUIC::Uint& maxRank, QUIC::MatAllocator<Float> * alloc = &cpuAllocator);
	virtual ~Basis();
	iterator begin();
	iterator end();
	const QUIC::Uint& size() const;
	void project(QUIC::Matrix<Float> * A, Float * result, Float * scratch, const QUIC::Uint& pitch);
	virtual void orthogonalizeRows(iterator newRowIters [], Float * scratch, const QUIC::Uint& scratchPitch);
	virtual void deleteRow(CTNode<Float> * n);
	virtual void copyRow(const iterator& source, const iterator& dest);
	void getMatrix(QUIC::Matrix<Float>& m);
	iterator createNewRow();
	static QUIC::Uint mgsBlockSize;
protected:
	virtual void realloc();
	Float * row(const QUIC::Uint& i);
	std::list<QUIC::Uint> fullList;
	QUIC::Uint pitch;
	QUIC::Uint lastRowRemoved;
	QUIC::Uint maxRows;
	Float * data;
	QUIC::Uint nCols;
	QUIC::Uint nRows;
	const QUIC::OrthoMode orthoMode;
	static const QUIC::Uint rowPreallocDenom = ROW_PREALLOC_DENOM;
	static const QUIC::Uint noRow;
	QUIC::MatAllocator<Float> * allocator;
	static QUIC::MatAllocator<Float> cpuAllocator;
};

template <typename Float>
class BasisGPU : public Basis<Float>
{
public:
	BasisGPU(const QUIC::Uint& m, const QUIC::Uint& n, const QUIC::OrthoMode& om, const QUIC::Uint& maxRank);
	~BasisGPU();
	void project(QUIC::Matrix<Float> * A, Float * result, Float * scratch, const QUIC::Uint& pitch);
	void orthogonalizeRows(Basis<Float>::iterator newRowIters [], Float * scratch, const QUIC::Uint& scratchPitch);
	void copyRow(const Basis<Float>::iterator& source, const Basis<Float>::iterator& dest);
private:
	void realloc();
	static QUIC::MatAllocatorGPU<Float> gpuAllocator;
};

template <typename Float>
inline Basis<Float>::iterator::iterator(const std::list<QUIC::Uint>::iterator& i, Basis<Float> * b)
: std::list<QUIC::Uint>::iterator(i), basis(b)
{}

template <typename Float>
inline typename Basis<Float>::iterator& Basis<Float>::iterator::operator++()
{
	std::list<QUIC::Uint>::iterator::operator++();
	return *this;
}

template <typename Float>
inline typename Basis<Float>::iterator Basis<Float>::iterator::operator++(int)
{
	Basis<Float>::iterator temp = *this;
	std::list<QUIC::Uint>::iterator::operator++();
	return temp;
}

template <typename Float>
inline Float * Basis<Float>::iterator::operator*() const
{
	return basis->row(std::list<QUIC::Uint>::iterator::operator*());
}

template <typename Float>
inline QUIC::Uint Basis<Float>::iterator::index() const
{
	return std::list<QUIC::Uint>::iterator::operator*();
}

template <typename Float>
inline typename Basis<Float>::iterator Basis<Float>::begin()
{
	return iterator(fullList.begin(), this);
}

template <typename Float>
inline typename Basis<Float>::iterator Basis<Float>::end()
{
	return iterator(fullList.end(), this);
}

template <typename Float>
inline const QUIC::Uint& Basis<Float>::size() const
{
	return nRows;
}

template <typename Float>
inline void Basis<Float>::getMatrix(QUIC::Matrix<Float>& m)
{
	m.setUsrData(nRows, nCols, data, pitch);
}

template <typename Float>
inline Float * Basis<Float>::row(const QUIC::Uint& i)
{
	return arrayRow(data, i, pitch);
}