#include "CTNode.h"
#include "CosineTree.h"
#include "globalFuncs.h"
#include "matrix.h"
#include "profile.h"
#include "../debugPrint.h"

//---------------------------------------------------------------------------------------------------------------

// CTNode class

template <typename Float>
CTNode<Float>::CTNode(const QUIC::Uint& nr, CosineTree<Float> * ct, const typename Basis<Float>::iterator& iter)
: rows(QUIC_SAFE_ALLOC(QUIC::Uint, nr)), nRows(nr), cosineTree(ct), basisVector(QUIC_SAFE_NEW(Basis<Float>::iterator, iter)), children(NULL)
{
	assert(nRows > 0);
}

template <typename Float>
CTNode<Float>::CTNode(const QUIC::Uint& nr, CosineTree<Float> * ct)
: rows(QUIC_SAFE_ALLOC(QUIC::Uint, nr)), nRows(nr), cosineTree(ct), basisVector(NULL), children(NULL)
{
	assert(nRows > 0);
}

template <typename Float>
CTNode<Float>::~CTNode()
{
	if (rows) delete [] rows;
	if (children)
	{
		(children + LEFT)->~CTNode();
		(children + RIGHT)->~CTNode();
		hostFree(children);
	}
	if (basisVector) delete basisVector;
}

template <typename Float>
CTNode<Float> * CTNode<Float>::split()
{
	if (nRows == 1) return NULL;
	LSDistribution<Float>::RowSampler rs(this);
	QUIC::Uint split = rs();
	CosineTree<Float>& ct = *cosineTree;
	const QUIC::Matrix<Float>& A = *ct.A;
	Float * const cosines = ct.scratchRow<Float>(0);
	QUIC::Uint * const labels = ct.scratchRow<QUIC::Uint>(1);
	assert(split < A.nRows);
	const Float * splitRow = ct.row(split);
	Float splitRowNorm = ct.ARowNorms[split];
	Float cMax = 0.0f;
	Float cMin = 1.0f;
	START_TIMER(COSINES, CPU);
	for (QUIC::Uint i=0; i<nRows; ++i)
	{
		const QUIC::Uint& rowId = rows[i];
		const Float * row = ct.row(rowId);
		const Float& rowNorm = ct.ARowNorms[rowId];
		Float dotProd = cpuDot(A.nCols, row, splitRow);
		Float& cosine = cosines[i];
		cosine = Abs(dotProd/(rowNorm*splitRowNorm));
		if (cosine > cMax) cMax = cosine;
		if (cosine < cMin) cMin = cosine;
	}
	STOP_TIMER(COSINES, CPU);
	if (cMin >= MAX_COS) return NULL;

	QUIC::Uint blockSize, nBlocks;
	cpuLabelRows(cosines, nRows, cMin, cMax, labels, blockSize, nBlocks);
	QUIC::Uint leftSize = countBits(labels, nBlocks);
	QUIC::Uint rightSize = nRows - leftSize;

	if (leftSize == 0 || rightSize == 0) return NULL;

	if (basisVector) ct.commonVars.basis->deleteRow(this);
	Basis<Float>& basis = *ct.commonVars.basis;

	Basis<Float>::iterator newRows [] = {basis.createNewRow(), basis.createNewRow()};

	children = QUIC_SAFE_ALLOC(CTNode, 2);
	new (children + LEFT) CTNode(leftSize, cosineTree, newRows[LEFT]);
	new (children + RIGHT) CTNode(rightSize, cosineTree, newRows[RIGHT]);

	populateChildRows(labels, blockSize, nBlocks);

	START_TIMER(ROW_MEAN, CPU);
	children[LEFT].getBasisVector(*newRows[LEFT]);
	children[RIGHT].getBasisVector(*newRows[RIGHT]);
	STOP_TIMER(ROW_MEAN, CPU);

	basis.orthogonalizeRows(newRows, ct.scratchRow<Float>(0), ct.getScratchPitch());

	delete [] rows;
	rows = NULL;
	nRows = 0;
	return children;
}

template <typename Float>
void CTNode<Float>::getBasisVector(Float * c)
{
	CosineTree<Float>& ct = *cosineTree;
	const QUIC::Matrix<Float>& A = *ct.A;
	cpuMemcpy(A.nCols, A.row(rows[0]), c);
	for (QUIC::Uint i=1; i<nRows; ++i) cpuVectorSum(A.nCols, 1.0f, A.row(rows[i]), c);
	Float norm = cpuNorm(A.nCols, c);
	cpuScalarMul(A.nCols, 1.0f/norm, c);
}

//---------------------------------------------------------------------------------------------------------------

// CTNodeGPU class

template <typename Float>
CTNodeGPU<Float>::CTNodeGPU(const QUIC::Uint& nr, CosineTree<Float> * ct, const typename Basis<Float>::iterator& iter)
: CTNode(nr, ct, iter)
{}

template <typename Float>
CTNodeGPU<Float>::CTNodeGPU(const QUIC::Uint& nr, CosineTree<Float> * ct)
: CTNode(nr, ct)
{}

template <typename Float>
CTNodeGPU<Float>::~CTNodeGPU()
{}

template <typename Float>
CTNode<Float> * CTNodeGPU<Float>::split()
{
	if (nRows == 1)
	{
		DEBUG_PRINT("Split aborted (%s, line %d\n", __FILE__, __LINE__);
		return NULL;
	}
	LSDistribution<Float>::RowSampler rs(this);
	QUIC::Uint split = rs();
	CosineTreeGPU<Float>& ct = *dynamic_cast<CosineTreeGPU<Float> *>(cosineTree);
	const QUIC::MatrixGPU<Float>& A = *dynamic_cast<const QUIC::MatrixGPU<Float> *>(ct.A);
	Float * const cosines = ct.scratchRow<Float>(1);
	QUIC::Uint * const labels = ct.scratchRow<QUIC::Uint>(0);
	assert(split < A.nRows);
	const Float * splitRow = ct.row(split);
	START_TIMER(COSINES, GPU);
	Float cMin, cMax;
	START_TIMER(COPY_H2D, GPU);
	if (!QUIC_SAFE_CALL(cudaMemcpy(ct.commonVarsGPU.rowIndexBuffer, rows, nRows*sizeof(QUIC::Uint), cudaMemcpyHostToDevice)))
	{
		DEBUG_PRINT("Split aborted (%s, line %d\n", __FILE__, __LINE__);
		return NULL;
	}
	STOP_TIMER(COPY_H2D, GPU);
	gpuCosines(A, ct.ARowNorms, ct.commonVarsGPU.rowIndexBuffer, nRows, split, cosines, cMin, cMax, ct.scratchRow<Float>(0), ct.scratchRow<Float>(2));
	STOP_TIMER(COSINES, GPU);
	if (cMin > MAX_COS)
	{
		DEBUG_PRINT("Split aborted (%s, line %d\n", __FILE__, __LINE__);
		return NULL;
	}

	QUIC::Uint blockSize, nBlocks;
	gpuLabelRows(cosines, nRows, cMin, cMax, labels, blockSize, nBlocks);
	QUIC::Uint * hLabels = reinterpret_cast<QUIC::Uint *>(ct.commonVars.sampleProjBuffer);
	START_TIMER(COPY_D2H, GPU);
	if (!QUIC_SAFE_CALL(cudaMemcpy(hLabels, labels, nBlocks*sizeof(QUIC::Uint), cudaMemcpyDeviceToHost))) 
	{
		DEBUG_PRINT("Split aborted (%s, line %d\n", __FILE__, __LINE__);
		return NULL;
	}
	STOP_TIMER(COPY_D2H, GPU);
	QUIC::Uint leftSize = countBits(hLabels, nBlocks);
	QUIC::Uint rightSize = nRows - leftSize;

	if (leftSize == 0 || rightSize == 0) 
	{
		DEBUG_PRINT("Split aborted (%s, line %d\n", __FILE__, __LINE__);
		return NULL;
	}

	if (basisVector) ct.commonVars.basis->deleteRow(this);
	BasisGPU<Float>& basis = *dynamic_cast<BasisGPU<Float> *>(ct.commonVars.basis);
	
	BasisGPU<Float>::iterator newRows [] = {basis.createNewRow(), basis.createNewRow()};

	CTNodeGPU * childrenGPU = QUIC_SAFE_ALLOC(CTNodeGPU, 2);
	new (childrenGPU + LEFT) CTNodeGPU(leftSize, cosineTree, newRows[LEFT]);
	new (childrenGPU + RIGHT) CTNodeGPU(rightSize, cosineTree, newRows[RIGHT]);
	children = childrenGPU;

	populateChildRows(hLabels, blockSize, nBlocks);
	
	START_TIMER(ROW_MEAN, GPU);
	gpuRowMeans(A, ct.commonVarsGPU.rowIndexBuffer, nRows, labels, *newRows[LEFT], *newRows[RIGHT], ct.scratchRow<Float>(1), ct.getScratchPitch());
	STOP_TIMER(ROW_MEAN, GPU);

	basis.orthogonalizeRows(newRows, ct.scratchRow<Float>(0), ct.getScratchPitch());

	delete [] rows;
	rows = NULL;
	nRows = 0;
	DEBUG_PRINT("Split successful (%s, line %d\n", __FILE__, __LINE__);
	return children;
}

template <typename Float>
void CTNodeGPU<Float>::getBasisVector(Float * c)
{
	CosineTree<Float>& ct = *cosineTree;
	const QUIC::MatrixGPU<Float>& A = *dynamic_cast<const QUIC::MatrixGPU<Float> *>(ct.A);
	gpuMemcpy(A.nCols, A.row(rows[0]), c);
	for (QUIC::Uint i=1; i<nRows; ++i) gpuVectorSum(A.nCols, 1.0f, ct.row(rows[i]), c);
	Float norm = gpuNorm(A.nCols, c);
	gpuScalarMul(A.nCols, 1.0f/norm, c);
}

//---------------------------------------------------------------------------------------------------------------

// explicit instantiations

INSTANTIATE_CLASS_TEMPLATE(CTNode)