#include "CosineTree.h"
#include "globalFuncs.h"
#include "Matrix.h"
#include "profile.h"

//---------------------------------------------------------------------------------------------------------------

// CommonVars nested class

template <typename Float>
CosineTree<Float>::CommonVars::CommonVars()
{}

template <typename Float>
CosineTree<Float>::CommonVars::CommonVars(const QUIC::Matrix<Float> * A, const QUIC::Uint& ms, const QUIC::OrthoMode& om, const QUIC::Uint& maxRank, CosineTree<Float> * ct)
: scratchW(std::max<QUIC::Uint>(A->nRows, A->nCols)*sizeof(Float)), scratchH(std::max<QUIC::Uint>(ms, 3)), maxSubsetSize(ms), Asubset(QUIC_SAFE_NEW(QUIC::Matrix<Float>, ms, A->nCols)),
  sampleIndexBuffer(QUIC_SAFE_ALLOC(QUIC::Uint, ms)), sampleProjBuffer(QUIC_SAFE_ALLOC(Float, ms)), basis(QUIC_SAFE_NEW(Basis<Float>, A->nRows, A->nCols, om, maxRank)), owner(ct)
{
	scratch = QUIC_SAFE_ALLOC(char, scratchW*scratchH);
	scratchPitch = scratchW;
}

template <typename Float>
CosineTree<Float>::CommonVars::CommonVars(const QUIC::MatrixGPU<Float> * A, const QUIC::Uint& ms, const QUIC::OrthoMode& om, const QUIC::Uint& maxRank, CosineTree<Float> * ct)
: scratchW(std::max<QUIC::Uint>(A->nRows, A->nCols)*sizeof(Float)), scratchH(std::max<QUIC::Uint>(ms, 3)), maxSubsetSize(ms), Asubset(QUIC_SAFE_NEW(QUIC::MatrixGPU<Float>, ms, A->nCols)),
  sampleIndexBuffer(QUIC_SAFE_ALLOC(QUIC::Uint, ms)), sampleProjBuffer(hostAllocMapped<Float>(std::max(A->nRows, A->nCols))), basis(QUIC_SAFE_NEW(BasisGPU<Float>, A->nRows, A->nCols, om, maxRank)), owner(ct)
{
	scratch = devAlloc2D<char>(scratchW, scratchH, scratchPitch);
}

//---------------------------------------------------------------------------------------------------------------

// CommonVarsGPU nested class

template <typename Float>
CosineTreeGPU<Float>::CommonVarsGPU::CommonVarsGPU()
{}

template <typename Float>
CosineTreeGPU<Float>::CommonVarsGPU::CommonVarsGPU(const QUIC::MatrixGPU<Float> * A, const CosineTree<Float>::CommonVars& cv, CosineTreeGPU<Float> * ct)
: rowIndexBuffer(devAlloc<QUIC::Uint>(A->nRows)), owner(ct)
{
	QUIC_SAFE_CALL(cudaHostGetDevicePointer(reinterpret_cast<void **>(&dSampleProjBuffer), cv.sampleProjBuffer, 0));
}

//---------------------------------------------------------------------------------------------------------------

// CosineTree class

template <typename Float>
CosineTree<Float>::CosineTree(const QUIC::MatrixGPU<Float> * m, const QUIC::OrthoMode& om, const QUIC::Uint& maxRank, CosineTree * ct)
: A(m), LS(m), ARowNorms(devAlloc<Float>(m->nRows))
{
	START_TIMER(ROW_NORMS, GPU);
	rowNorms(*m, false, ARowNorms);
	STOP_TIMER(ROW_NORMS, GPU);
	Float * temp = QUIC_SAFE_ALLOC(Float, A->nRows);
	QUIC_SAFE_CALL(cudaMemcpy(temp, ARowNorms, A->nRows*sizeof(Float), cudaMemcpyDeviceToHost));
	LS.initDistribution(temp);
	hostFree(temp);
	if (ct)
	{
		commonVars = ct->commonVars;
		root = QUIC_SAFE_NEW(CTNodeGPU<Float>, m->nRows, this);
		root->initRows();
	}
	else
	{
		commonVars = CommonVars(m, nSamples(m->nRows), om, maxRank, this);
		BasisGPU<Float>::iterator iter = commonVars.basis->createNewRow();
		root = QUIC_SAFE_NEW(CTNodeGPU<Float>, m->nRows, this, iter);
		root->initRows();
		root->getBasisVector(*iter);
	}
	nASq = LS.normSq();
}

template <typename Float>
CosineTree<Float>::CosineTree(const QUIC::Matrix<Float> * m, const QUIC::OrthoMode& om, const QUIC::Uint& maxRank, CosineTree * ct)
: A(m), LS(m), ARowNorms(QUIC_SAFE_ALLOC(Float, m->nRows))
{
	START_TIMER(ROW_NORMS, CPU);
	rowNorms(*m, false, ARowNorms);
	STOP_TIMER(ROW_NORMS, CPU);
	LS.initDistribution(ARowNorms);
	if (ct)
	{
		commonVars = ct->commonVars;
		root = QUIC_SAFE_NEW(CTNode<Float>, m->nRows, this);
		root->initRows();
	}
	else
	{
		commonVars = CommonVars(m, nSamples(m->nRows), om, maxRank, this);
		Basis<Float>::iterator iter = commonVars.basis->createNewRow();
		root = QUIC_SAFE_NEW(CTNode<Float>, m->nRows, this, iter);
		root->initRows();
		root->getBasisVector(*iter);
	}
	nASq = LS.normSq();
}

template <typename Float>
CosineTree<Float>::~CosineTree()
{
	delete root;
	if (commonVars.owner == this)
	{
		hostFree(commonVars.sampleIndexBuffer);
		delete commonVars.Asubset;
		if (ARowNorms) hostFree(ARowNorms);
		if (commonVars.scratch) hostFree(commonVars.scratch);
		if (commonVars.sampleProjBuffer) hostFree(commonVars.sampleProjBuffer);
		delete commonVars.basis;
	}
}

template <typename Float>
bool CosineTree<Float>::extractSVD(QUIC::Matrix<Float> * U, QUIC::Matrix<Float> * S, QUIC::Matrix<Float> * VT)
{
	const QUIC::Uint nBasisRows = commonVars.basis->size();
	const QUIC::Uint& nRows = A->nRows;
	const QUIC::Uint& nCols = A->nCols;
	QUIC::Matrix<Float>& Sr = *S;
	QUIC::Matrix<Float> VhatT;
	commonVars.basis->getMatrix(VhatT);
	QUIC::Matrix<Float> AVhat(nRows, VhatT.nRows);
	QUIC::Matrix<Float>::mul(*A, VhatT, AVhat, false, true);
	return extractSVDFromProjection<QUIC::Matrix<Float>, Float>(AVhat, VhatT, U, S, VT, scratchRow<Float>(0), commonVars.scratchPitch);
}

template <typename Float>
void CosineTree<Float>::basisProject()
{
	commonVars.basis->project(commonVars.Asubset, commonVars.sampleProjBuffer, scratchRow<Float>(0), commonVars.scratchPitch);
}

//---------------------------------------------------------------------------------------------------------------

// CosineGPU class

template <typename Float>
CosineTreeGPU<Float>::CosineTreeGPU(const QUIC::MatrixGPU<Float> * m, const QUIC::OrthoMode& om, const QUIC::Uint& maxRank, CosineTreeGPU * ct)
: CosineTree(m, om, maxRank, ct)
{
	if (ct) commonVarsGPU = ct->commonVarsGPU;
	else commonVarsGPU = CommonVarsGPU(m, commonVars, this);
}

template <typename Float>
CosineTreeGPU<Float>::~CosineTreeGPU()
{
	if (commonVars.owner == this)
	{
		QUIC_SAFE_CALL(cudaFreeHost(commonVars.sampleProjBuffer));
		commonVars.sampleProjBuffer = NULL;
		devFree(commonVars.scratch);
		commonVars.scratch = NULL;
	}
	devFree(ARowNorms);
	ARowNorms = NULL;
	if (commonVarsGPU.owner == this)
	{
		devFree(commonVarsGPU.rowIndexBuffer);
	}
}

template <typename Float>
bool CosineTreeGPU<Float>::extractSVD(QUIC::Matrix<Float> * U, QUIC::Matrix<Float> * S, QUIC::Matrix<Float> * VT)
{
	const QUIC::Uint nBasisRows = commonVars.basis->size();
	const QUIC::Uint& nRows = A->nRows;
	const QUIC::Uint& nCols = A->nCols;
	const QUIC::MatrixGPU<Float>& Ar = *dynamic_cast<const QUIC::MatrixGPU<Float> *>(A);
	QUIC::MatrixGPU<Float>& Sr = *dynamic_cast<QUIC::MatrixGPU<Float> *>(S);
	QUIC::MatrixGPU<Float> VhatT;
	commonVars.basis->getMatrix(VhatT);
	QUIC::MatrixGPU<Float> AVhat(nRows, VhatT.nRows);
	QUIC::MatrixGPU<Float>::mul(Ar, VhatT, AVhat, false, true);
	QUIC::MatrixGPU<Float> * Ugpu = U ? dynamic_cast<QUIC::MatrixGPU<Float> *>(U) : NULL;
	QUIC::MatrixGPU<Float> * Sgpu = dynamic_cast<QUIC::MatrixGPU<Float> *>(S);
	QUIC::MatrixGPU<Float> * VTgpu = VT ? dynamic_cast<QUIC::MatrixGPU<Float> *>(VT) : NULL;
	return extractSVDFromProjection<QUIC::MatrixGPU<Float>, Float>(AVhat, VhatT, Ugpu, Sgpu, VTgpu, scratchRow<Float>(0), commonVars.scratchPitch);
}

template <typename Float>
void CosineTreeGPU<Float>::basisProject()
{
	dynamic_cast<BasisGPU<Float> *>(commonVars.basis)->project(commonVars.Asubset, commonVarsGPU.dSampleProjBuffer, scratchRow<Float>(0), commonVars.scratchPitch);
}

//---------------------------------------------------------------------------------------------------------------

// explicit instantiations

INSTANTIATE_CLASS_TEMPLATE(CosineTree)