#pragma once

#include <list>
#include "Basis.h"
#include "CTNode.h"
#include "../debugPrint.h"

#define MAX_COS 0.9999999999

template <typename Float>
class CTNode;

template <typename Float>
class CosineTree : public QUIC::Base
{
public:
	class CommonVars
	{
	public:
		CommonVars();
		CommonVars(const QUIC::Matrix<Float> * A, const QUIC::Uint& ms, const QUIC::OrthoMode& om, const QUIC::Uint& maxRank, CosineTree * ct);
		CommonVars(const QUIC::MatrixGPU<Float> * A, const QUIC::Uint& ms, const QUIC::OrthoMode& om, const QUIC::Uint& maxRank, CosineTree * ct);
		template <typename T>
		T * scratchRow(const QUIC::Uint& i);
		QUIC::Uint maxSubsetSize;
		char * scratch;
		QUIC::Uint scratchW;
		QUIC::Uint scratchH;
		QUIC::Uint scratchPitch;
		QUIC::Matrix<Float> * Asubset;
		QUIC::Uint * sampleIndexBuffer;
		Float * sampleProjBuffer;
		CosineTree<Float> * owner;
		Basis<Float> * basis;
	};
	typedef CTNodeGPU<Float> Node;
	typedef QUIC::Matrix<Float> Matrix;
	CosineTree(const QUIC::MatrixGPU<Float> * m, const QUIC::OrthoMode& om, const QUIC::Uint& maxRank, CosineTree * ct);
	CosineTree(const QUIC::Matrix<Float> * m, const QUIC::OrthoMode& om, const QUIC::Uint& maxRank, CosineTree * ct);
	virtual ~CosineTree();
	virtual bool extractSVD(QUIC::Matrix<Float> * U, QUIC::Matrix<Float> * S, QUIC::Matrix<Float> * VT);
	virtual void basisProject();
	Float MCSqError(const typename LSDistribution<Float>::RowSampler& rs, const Float& delta, const QUIC::Uint& s);
	const Float * row(const QUIC::Uint& r);
	template <typename T>
	T * scratchRow(const QUIC::Uint& i);
	const QUIC::Uint& getScratchPitch() const;
	double nASq;
	CTNode<Float> * root;
	Float * ARowNorms;
	const QUIC::Matrix<Float> * A;
	LSDistribution<Float> LS;
	CommonVars commonVars;
};

template <typename Float>
class CosineTreeGPU : public CosineTree<Float>
{
public:
	class CommonVarsGPU
	{
		friend class CosineTreeGPU<Float>;
	public:
		CommonVarsGPU();
		CommonVarsGPU(const QUIC::MatrixGPU<Float> * A, const CommonVars& cv, CosineTreeGPU * ct);
		QUIC::Uint * rowIndexBuffer;
	private:
		Float * dSampleProjBuffer;
		CosineTreeGPU<Float> * owner;
	};
	typedef CTNodeGPU<Float>  Node;
	typedef QUIC::MatrixGPU<Float> Matrix;
	CosineTreeGPU(const QUIC::MatrixGPU<Float> * m, const QUIC::OrthoMode& om, const QUIC::Uint& maxRank, CosineTreeGPU * ct);
	~CosineTreeGPU();
	bool extractSVD(QUIC::Matrix<Float> * U, QUIC::Matrix<Float> * S, QUIC::Matrix<Float> * VT);
	void basisProject();
	CommonVarsGPU commonVarsGPU;
};

template <typename Float>
template <typename T>
inline T * CosineTree<Float>::CommonVars::scratchRow(const QUIC::Uint& i)
{
	assert(i < scratchH);
	return reinterpret_cast<T *>(scratch + i*scratchPitch);
}

template <typename Float>
inline const Float * CosineTree<Float>::row(const QUIC::Uint& r)
{
	return A->row(r);
}

template <typename Float>
template <typename T>
inline T * CosineTree<Float>::scratchRow(const QUIC::Uint& i)
{
	return commonVars.scratchRow<T>(i);
}

template <typename Float>
inline const QUIC::Uint& CosineTree<Float>::getScratchPitch() const
{
	return commonVars.scratchPitch;
}

template <typename Float>
inline Float CosineTree<Float>::MCSqError(const typename LSDistribution<Float>::RowSampler& rs, const Float& delta, const QUIC::Uint& s)
{
	rs(s, commonVars.sampleIndexBuffer);
	commonVars.Asubset->nRows = s;
	if (commonVars.scratchH < commonVars.Asubset->nRows) throw (QUICException(__FILE__, __LINE__, "Dimension mismatch"));
	A->copySubset(commonVars.sampleIndexBuffer, commonVars.Asubset);
	basisProject();
	Float mean = 0.0f;
	Float n = static_cast<Float>(s);
	QUIC::Uint * r = commonVars.sampleIndexBuffer;
	for (QUIC::Uint i=0; i<s; ++i)
	{
		Float& p = commonVars.sampleProjBuffer[i];
		p /= static_cast<Float>(rs[*(r++)]);;
		mean += p;
	}
	mean /= n;
	Float variance = 0.0f;
	for (QUIC::Uint i=0; i<s; ++i)
	{
		Float d = commonVars.sampleProjBuffer[i] - mean;
		variance += d*d;
	}
	variance /= n;
	if (mean < 0.0 || variance < 0.0) throw (QUICException(__FILE__, __LINE__, "Negative mean or variance"));
	double e = icdf(delta, mean, variance);
	Float result = static_cast<Float>(nASq - e);
	if (result < 0.0) throw (QUICException(__FILE__, __LINE__, "Negative error estimate"));
	return result;
}