#pragma once

#include "QUIC.h"
#include "globalFuncs.h"

template <typename Float>
class CTNode;

template <typename Float>
class LSDistribution : public QUIC::Base
{
public:
	class RowSampler : public QUIC::Base
	{
	public:
		RowSampler(CTNode<Float> * n);
		~RowSampler();
		QUIC::Uint operator()() const;
		void operator()(const QUIC::Uint& n, QUIC::Uint * samples) const;
		double operator[](const QUIC::Uint& i) const;
		static const double * binarySearch(const double * r, const QUIC::Uint& nr, const double& x);
		QUIC::Uint sample(const QUIC::Uint& seed) const;
		const double& sum() const;
	private:
		double * prefixSums;
		const LSDistribution * LS;
		double distributionSum;
		QUIC::Uint nRows;
		CTNode<Float> * N;
	};
	const double& operator[](const QUIC::Uint& i) const;
	LSDistribution();
	LSDistribution(const QUIC::Matrix<Float> * A);
	~LSDistribution();
	void initDistribution(Float * rowNorms);
	double normSq() const;
private:
	double * distribution;
	QUIC::Uint n;
	double nA;
};

template <typename Float>
LSDistribution<Float>::RowSampler::RowSampler(CTNode<Float> * n)
: LS(&(n->LS())), prefixSums(QUIC_SAFE_ALLOC(double, n->size() + 1)), nRows(n->size()), N(n)
{
	assert(n > 0);
	prefixSums[0] = 0.0;
	for (QUIC::Uint i=0; i<nRows; ++i) prefixSums[i + 1] = prefixSums[i] + LS->distribution[(*n)[i]];
	distributionSum = prefixSums[nRows];
}

template <typename Float>
LSDistribution<Float>::RowSampler::~RowSampler()
{
	delete [] prefixSums;
}

template <typename Float>
QUIC::Uint LSDistribution<Float>::RowSampler::operator()() const
{
	return sample(rand());
}

template <typename Float>
void LSDistribution<Float>::RowSampler::operator()(const QUIC::Uint& n, QUIC::Uint * samples) const
{
	for (QUIC::Uint i=0; i<n; ++i) samples[i] = rand(); // serialize random number generation to avoid repeat rows
	#pragma omp parallel for
	for (QUIC::Int i=0; i<n; ++i) samples[i] = sample(samples[i]);
}

template <typename Float>
inline double LSDistribution<Float>::RowSampler::operator[](const QUIC::Uint& i) const
{
	return (*LS)[i]/distributionSum;
}

template <typename Float>
inline const double * LSDistribution<Float>::RowSampler::binarySearch(const double * r, const QUIC::Uint& nr, const double& x)
{
	const double * lb = r;
	const double * ub = r + nr;
	while (ub - lb > 1)
	{
		const double * m = lb + (ub - lb)/2;
		if (x > *m) lb = m;
		else if (x < *m) ub = m;
		else return m;
	}
	return ub;
}

template <typename Float>
inline QUIC::Uint LSDistribution<Float>::RowSampler::sample(const QUIC::Uint& seed) const
{
	assert(nRows > 0);
	CTNode<Float>& node = *N;
	double r = distributionSum*static_cast<double>(seed)/static_cast<double>(RAND_MAX);
	const double * p = binarySearch(prefixSums, nRows, r);
	QUIC::Uint i = static_cast<QUIC::Uint>(p - prefixSums);
	assert(i > 0);
	assert(i <= nRows);
	assert(prefixSums[i - 1] <= r);
	assert(prefixSums[i] >= r);
	assert(node.size() == 0 || i - 1 < node.size());
	return node.size() ? node[i - 1] : i - 1;
}

template <typename Float>
inline const double& LSDistribution<Float>::RowSampler::sum() const
{
	return distributionSum;
}

template <typename Float>
LSDistribution<Float>::LSDistribution()
: distribution(NULL), n(0), nA(0.0)
{}

template <typename Float>
LSDistribution<Float>::LSDistribution(const QUIC::Matrix<Float> * A)
: distribution(QUIC_SAFE_ALLOC(double, A->nRows)), n(A->nRows), nA(0.0f)
{}

template <typename Float>
void LSDistribution<Float>::initDistribution(Float * rowNorms)
{
	double distributionSum = 0.0;
	for (QUIC::Uint r=0; r<n; ++r)
	{
		double norm = rowNorms[r];
		double& p = distribution[r];
		p = norm*norm;
		distributionSum += p;
	}
	nA = Sqrt(distributionSum);
	cpuScalarMul(n, 1.0f/distributionSum, distribution);
}


template <typename Float>
LSDistribution<Float>::~LSDistribution()
{
	if (distribution) delete [] distribution;
}

template <typename Float>
inline const double& LSDistribution<Float>::operator[](const QUIC::Uint& i) const
{
	return distribution[i];
}

template <typename Float>
inline double LSDistribution<Float>::normSq() const
{
	return nA*nA;
}