#include "Matrix.h"
#include "globalFuncs.h"
#include "profile.h"

//---------------------------------------------------------------------------------------------------------------

// Matrix class

template <typename Float>
QUIC::Matrix<Float>::Matrix(std::istream& ist, QUIC::MatAllocator<Float> * alloc)
: nRows(0), nCols(0), data(NULL), usrData(false), allocator(alloc)
{
	std::string format;
	getline(ist, format, '\0');
	int nr, nc;
	ist.read(reinterpret_cast<char *>(&nr), sizeof(int));
	ist.read(reinterpret_cast<char *>(&nc), sizeof(int));
	nRows = static_cast<QUIC::Uint>(nr);
	nCols = static_cast<QUIC::Uint>(nc);
	QUIC::MatAllocator<Float>::pitched_pointer p = allocator->allocate(nCols, nRows);
	data = p.addr;
	pitch = p.pitch;
	if (nRows && nCols)
	{
		if (format == "float32") readFile<float>(ist);
		else if (format == "float64") readFile<double>(ist);
		else throw(QUICException(__FILE__, __LINE__, "Invalid data type"));
	}
}

template <typename Float>
QUIC::Matrix<Float>::Matrix(const QUIC::Uint& nr, const QUIC::Uint& nc, Float * d, const QUIC::Uint& p, QUIC::MatAllocator<Float> * alloc)
: nRows(nr), nCols(nc), data(d), pitch(p), usrData(true), allocator(alloc)
{}

template <typename Float>
QUIC::Matrix<Float>::Matrix(const QUIC::Uint& nr, const QUIC::Uint& nc, QUIC::MatAllocator<Float> * alloc)
: nRows(nr && nc ? nr : 0), nCols(nr && nc ? nc : 0), data(NULL), usrData(false), allocator(alloc)
{
	QUIC::MatAllocator<Float>::pitched_pointer p = allocator->allocate(nCols, nRows);
	data = p.addr;
	pitch = p.pitch;
}

template <typename Float>
QUIC::Matrix<Float>::Matrix(const Matrix& m, QUIC::MatAllocator<Float> * alloc)
: nRows(m.nRows), nCols(m.nCols), data(NULL), usrData(false), allocator(alloc)
{
	QUIC::MatAllocator<Float>::pitched_pointer p = allocator->allocate(nCols, nRows);
	data = p.addr;
	pitch = p.pitch;
	if (nRows && nCols)
	{
		#pragma omp parallel for
		for (QUIC::Int i=0; i<nRows; ++i) cpuMemcpy(nCols, m.row(i), row(i));
	}
}

template <typename Float>
QUIC::Matrix<Float>::Matrix(const MatrixGPU<Float>& m, QUIC::MatAllocator<Float> * alloc)
: nRows(m.nRows), nCols(m.nCols), data(NULL), usrData(false), allocator(alloc)
{
	QUIC::MatAllocator<Float>::pitched_pointer p = allocator->allocate(nCols, nRows);
	data = p.addr;
	pitch = p.pitch;
	if (data) QUIC_SAFE_CALL(cudaMemcpy2D(data, pitch, m.data, m.pitch, nCols*sizeof(Float), nRows, cudaMemcpyDeviceToHost));
}

template <typename Float>
QUIC::Matrix<Float>::Matrix(QUIC::MatAllocator<Float> * alloc)
: nRows(0), nCols(0), data(NULL), pitch(0), usrData(false), allocator(alloc)
{}

template <typename Float>
QUIC::Matrix<Float>::~Matrix()
{
	if (!usrData && data)
	{
		QUIC::MatAllocator<Float>::pitched_pointer p = QUIC::MatAllocator<Float>::make_pitched_pointer(data, pitch);
		allocator->deallocate(p, nCols, nRows);
	}
}

template <typename Float>
QUIC::Matrix<Float>& QUIC::Matrix<Float>::operator=(const Matrix<Float>& m)
{
	if (nRows != m.nRows || nCols != m.nCols) realloc(m.nRows, m.nCols);
	if (data)
	{
		#pragma omp parallel for
		for (QUIC::Int i=0; i<nRows; ++i) cpuMemcpy(nCols, m.row(i), row(i));
	}
	return *this;
}

template <typename Float>
QUIC::Matrix<Float>& QUIC::Matrix<Float>::operator=(const MatrixGPU<Float>& m)
{
	if (nRows != m.nRows || nCols != m.nCols) realloc(m.nRows, m.nCols);
	if (data)
	{
		QUIC_SAFE_CALL(cudaMemcpy2D(data, pitch, m.data, m.pitch, nCols*sizeof(Float), nRows, cudaMemcpyDeviceToHost));
	}
	return *this;
}

template <typename Float>
QUIC::Matrix<Float> QUIC::Matrix<Float>::operator*(const Matrix<Float>& m) const
{
	Matrix<Float> result(nRows, m.nCols);
	Matrix<Float>::mul(*this, m, result, false, false);
	return result;
}

template <typename Float>
void QUIC::Matrix<Float>::realloc(const QUIC::Uint& nr, const QUIC::Uint& nc)
{
	QUIC::MatAllocator<Float>::pitched_pointer p = QUIC::MatAllocator<Float>::make_pitched_pointer(data, pitch);
	if (!usrData && data)
	{
		allocator->deallocate(p, nCols, nRows);
	}
	data = NULL;
	if (nr && nc) 
	{
		p = allocator->allocate(nc, nr);
		data = p.addr;
		pitch = p.pitch;
	}
	usrData = false;
	nRows = nr;
	nCols = nc;
}

template <typename Float>
void QUIC::Matrix<Float>::copySubset(const QUIC::Uint * subsetRows, QUIC::Matrix<Float> * A) const
{
	QUIC::Matrix<Float>& Ar = *A;
	QUIC::Uint nSubsetCols = std::min(nCols, Ar.nCols);
	#pragma omp parallel for
	for (QUIC::Int i=0; i<Ar.nRows; ++i)
	{
		const Float * sourceRow = row(subsetRows[i]);
		Float * destRow = Ar.row(i);
		cpuMemcpy(nSubsetCols, sourceRow, destRow);
	}
}

template <typename Float>
bool QUIC::Matrix<Float>::svd(QUIC::Matrix<Float>& U,QUIC::Matrix<Float>& S, QUIC::Matrix<Float>& VT) const
{
	U.realloc(nRows, nRows);
	S.realloc(1, nRows < nCols ? nRows : nCols);
	VT.realloc(nCols, nCols);
	QUIC::Uint lwork = 8*(nCols + nRows);
	Float * work = QUIC_SAFE_ALLOC(Float, lwork);
	QUIC::Uint m = nCols;
	QUIC::Uint n = nRows;
	QUIC::Uint ldA = pitch/sizeof(Float);
	QUIC::Uint ldU = U.pitch/sizeof(Float);
	QUIC::Uint ldVT = VT.pitch/sizeof(Float);
	bool rval = cpuSVD(CPU_FULL, CPU_FULL, m, n, data, ldA, S.data, VT.data, ldVT, U.data, ldU, work, lwork);
	delete [] work;
	return rval;
}

template <typename Float>
Float QUIC::Matrix<Float>::norm() const
{
	Float norm = 0.0f;
	for (QUIC::Uint i=0; i<nRows; ++i) norm += square(cpuNorm(nCols, row(i)));
	return Sqrt(norm);
}

template <typename Float>
void QUIC::Matrix<Float>::print(std::ostream& ost) const
{
	QUIC::Int * cw = new QUIC::Int [nCols];
	bool * cn = QUIC_SAFE_ALLOC(bool, nCols);
	std::string ** str = QUIC_SAFE_ALLOC(std::string *, nRows);
	for (QUIC::Uint i=0; i<nRows; ++i)
	{
		str[i] = new std::string [nCols];
	}
	for (QUIC::Uint i=0; i<nCols; ++i)
	{
		cw[i] = 0;
		cn[i] = false;
	}
	for (QUIC::Uint r=0; r<nRows; ++r)
	{
		const Float * p = row(r);
		for (QUIC::Uint c=0; c<nCols; ++c)
		{
			std::stringstream ss;
			ss<<*p;
			str[r][c] = ss.str();
			QUIC::Int s = static_cast<QUIC::Int>(str[r][c].size());
			if (*p < 0)
			{
				cn[c] = true;
				s -= 1;
			}
			if (s > cw[c]) cw[c] = s;
			++p;
		}
	}
	for (QUIC::Uint r=0; r<nRows; ++r)
	{
		for (QUIC::Uint c=0; c<nCols; ++c)
		{
			QUIC::Int s = static_cast<QUIC::Int>(str[r][c].size());
			if (str[r][c][0] == '-') --s;
			else if (cn[c]) ost<<' ';
			ost<<str[r][c];
			for (QUIC::Int i=s; i<=cw[c]; ++i) ost<<' ';
		}
		ost<<std::endl;
	}
	delete [] cw;
	delete [] cn;
	for (QUIC::Uint i=0; i<nRows; ++i) delete [] str[i];
	delete [] str;
}

template <typename Float>
template <typename FileType>
void QUIC::Matrix<Float>::readFile(std::istream& ist)
{
	FileType * temp = QUIC_SAFE_ALLOC(FileType, nRows*nCols);
	FileType * t = temp;
	ist.read(reinterpret_cast<char *>(temp), nRows*nCols*sizeof(FileType));
	if (ist.eof() || ist.fail()) throw (QUICException(__FILE__, __LINE__, "Failed to read file."));
	for (QUIC::Uint r=0; r<nRows; ++r)
	{
		Float * entry = row(r);
		for (QUIC::Uint c=0; c<nCols; ++c) *(entry++) = static_cast<Float>(*(t++));	
	}
	delete [] temp;
}

namespace QUIC
{
	template <>
	template <>
	void QUIC::Matrix<float>::readFile<float>(std::istream& ist)
	{
		for (QUIC::Uint i=0; i<nRows; ++i)
		{
			ist.read(reinterpret_cast<char *>(row(i)), nCols*sizeof(float));
			if (ist.eof() || ist.fail()) throw (QUICException(__FILE__, __LINE__, "Failed to read file."));
		}
	}

	template <>
	template <>
	void QUIC::Matrix<double>::readFile<double>(std::istream& ist)
	{
		for (QUIC::Uint i=0; i<nRows; ++i)
		{
			ist.read(reinterpret_cast<char *>(row(i)), nCols*sizeof(double));
			if (ist.eof() || ist.fail()) throw (QUICException(__FILE__, __LINE__, "Failed to read file."));
		}
	}
}

template <typename Float>
void QUIC::Matrix<Float>::mul(const Matrix& A, const Matrix& B, Matrix& C, bool AT, bool BT)
{
	START_TIMER(MUL, CPU);
	//A*B = C <==> BT*AT = CT
	QUIC::Uint opANRows, opANCols, opBNRows, opBNCols;
	getDim(A, opANRows, opANCols, !AT);
	getDim(B, opBNRows, opBNCols, !BT);
	//assert(opBNCols == opANRows);
	//assert(opBNRows == C.nCols);
	//assert(opANCols == C.nRows);
	CPU_TRANS_TYPE CBAT = AT ? CPU_TRANS : CPU_NO_TRANS;
	CPU_TRANS_TYPE CBBT = BT ? CPU_TRANS : CPU_NO_TRANS;
	QUIC::Uint ldA = A.pitch/sizeof(Float);
	QUIC::Uint ldB = B.pitch/sizeof(Float);
	QUIC::Uint ldC = C.pitch/sizeof(Float);
	cpuMatMul(CBBT, CBAT, opBNRows, opANCols, opBNCols, 1.0f, B.data, ldB, A.data, ldA, 0.0f, C.data, ldC);
	STOP_TIMER(MUL, CPU);
}

template <typename Float>
typename QUIC::MatAllocator<Float> QUIC::Matrix<Float>::cpuAllocator;

//---------------------------------------------------------------------------------------------------------------

// MatrixGPU class

template <typename Float>
QUIC::MatrixGPU<Float>::MatrixGPU(const QUIC::Uint& nr, const QUIC::Uint& nc)
: Matrix<Float>(nr, nc, &gpuAllocator)
{}

template <typename Float>
QUIC::MatrixGPU<Float>::MatrixGPU(const QUIC::Uint& nr, const QUIC::Uint& nc, Float * d, const QUIC::Uint& p)
: Matrix<Float>(nr, nc, d, p, &gpuAllocator)
{}

template <typename Float>
QUIC::MatrixGPU<Float>::MatrixGPU(const Matrix<Float>& m)
: Matrix<Float>(m.nRows, m.nCols, &gpuAllocator)
{
	usrData = false;
	if (nRows && nCols && !QUIC::checkError())
	{
		QUIC_SAFE_CALL(cudaMemcpy2D(data, pitch, m.data, m.pitch, m.nCols*sizeof(Float), m.nRows, cudaMemcpyHostToDevice));
	}
}

template <typename Float>
QUIC::MatrixGPU<Float>::MatrixGPU()
: Matrix<Float>(&gpuAllocator)
{}

template <typename Float>
QUIC::MatrixGPU<Float>::~MatrixGPU()
{}

template <typename Float>
QUIC::MatrixGPU<Float>& QUIC::MatrixGPU<Float>::operator=(const Matrix<Float>& m)
{
	if (nRows != m.nRows || nCols != m.nCols) realloc(m.nRows, m.nCols);
	if (data)
	{
		QUIC_SAFE_CALL(cudaMemcpy2D(data, pitch, m.data, m.pitch, nCols*sizeof(Float), nRows, cudaMemcpyHostToDevice));
	}
	return *this;
}

template <typename Float>
QUIC::MatrixGPU<Float>& QUIC::MatrixGPU<Float>::operator=(const MatrixGPU<Float>& m)
{
	if (nRows != m.nRows || nCols != m.nCols) realloc(m.nRows, m.nCols);
	if (data)
	{
		QUIC_SAFE_CALL(cudaMemcpy2D(data, pitch, m.data, m.pitch, nCols*sizeof(Float), nRows, cudaMemcpyDeviceToDevice));
	}
	return *this;
}

template <typename Float>
QUIC::MatrixGPU<Float> QUIC::MatrixGPU<Float>::operator*(const MatrixGPU<Float>& m) const
{
	MatrixGPU<Float> result(nRows, m.nCols);
	MatrixGPU<Float>::mul(*this, m, result, false, false);
	return result;
}

template <typename Float>
void QUIC::MatrixGPU<Float>::copySubset(const QUIC::Uint * subsetRows,  Matrix<Float> * A) const
{
	MatrixGPU& Ar = *dynamic_cast<MatrixGPU<Float> *>(A);
	QUIC::Uint rowWidth = std::min(nCols, Ar.nCols)*sizeof(Float);
	for (QUIC::Uint i=0; i<Ar.nRows; ++i)
	{
		const Float * sourceRow = row(subsetRows[i]);
		Float * destRow = Ar.row(i);
		QUIC_SAFE_CALL(cudaMemcpyAsync(destRow, sourceRow, rowWidth, cudaMemcpyDeviceToDevice));
	}
}

template <typename Float>
void QUIC::MatrixGPU<Float>::print(std::ostream& ost) const
{
	Matrix<Float> temp(*this);
	temp.print(ost);
}

template <typename Float>
bool QUIC::MatrixGPU<Float>::svd(MatrixGPU& U, MatrixGPU& S, MatrixGPU& VT) const
{
	U.realloc(nRows, nRows);
	S.realloc(1, nRows < nCols ? nRows : nCols);
	VT.realloc(nCols, nCols);
	QUIC::Uint ldA = pitch/sizeof(Float);
	QUIC::Uint ldU = U.pitch/sizeof(Float);
	QUIC::Uint ldVT = VT.pitch/sizeof(Float);
	return gpuSVD(GPU_FULL, GPU_FULL, nCols, nRows, data, ldA, S.data, VT.data, ldVT, U.data, ldU);
}

template <typename Float>
inline void getDim(const QUIC::Matrix<Float>& mat, QUIC::Uint& nRows, QUIC::Uint& nCols, bool transpose)
{
	if (transpose)
	{
		nRows = mat.nCols;
		nCols = mat.nRows;
	}
	else
	{
		nRows = mat.nRows;
		nCols = mat.nCols;
	}
}

template <typename Float>
void QUIC::MatrixGPU<Float>::mul(const MatrixGPU& A, const MatrixGPU& B, MatrixGPU& C, bool AT, bool BT)
{
	START_TIMER(MUL, GPU);
	QUIC::Uint opANRows, opANCols, opBNRows, opBNCols;
	getDim(A, opANRows, opANCols, !AT);
	getDim(B, opBNRows, opBNCols, !BT);
	//assert(opBNCols == opANRows);
	//assert(opBNRows == C.nCols);
	//assert(opANCols == C.nRows);
	GPU_TRANS_TYPE CBAT = AT ? GPU_TRANS : GPU_NO_TRANS;
	GPU_TRANS_TYPE CBBT = BT ? GPU_TRANS : GPU_NO_TRANS;
	QUIC::Uint ldA = A.pitch/sizeof(Float);
	QUIC::Uint ldB = B.pitch/sizeof(Float);
	QUIC::Uint ldC = C.pitch/sizeof(Float);
	gpuMatMul(CBBT, CBAT, opBNRows, opANCols, opBNCols, 1.0f, B.data, ldB, A.data, ldA, 0.0f, C.data, ldC);
	STOP_TIMER(MUL, GPU);
}

template <typename Float>
typename QUIC::MatAllocatorGPU<Float> QUIC::MatrixGPU<Float>::gpuAllocator;

//---------------------------------------------------------------------------------------------------------------

// explicit instantiations

INSTANTIATE_CLASS_TEMPLATE(QUIC::Matrix)