#pragma once

#include <cassert>
#include <fstream>
#include <sstream>
#include "QUIC.h"
#include "MatAllocator.h"

template <typename Float>
class QUIC::Matrix : public QUIC::Base
{
	friend class MatrixGPU<Float>;
public:
	Matrix(std::istream& ifs, QUIC::MatAllocator<Float> * alloc = &cpuAllocator);
	Matrix(const QUIC::Uint& nr, const QUIC::Uint& nc, Float * d, const QUIC::Uint& p, QUIC::MatAllocator<Float> * alloc = &cpuAllocator);
	Matrix(const QUIC::Uint& nr, const QUIC::Uint& nc, QUIC::MatAllocator<Float> * alloc = &cpuAllocator);
	Matrix(const Matrix& mp, QUIC::MatAllocator<Float> * alloc = &cpuAllocator);
	Matrix(const MatrixGPU<Float>& mp, QUIC::MatAllocator<Float> * alloc = &cpuAllocator);
	Matrix(QUIC::MatAllocator<Float> * alloc = &cpuAllocator);
	virtual ~Matrix();
	Matrix<Float>& operator=(const Matrix<Float>& m);
	Matrix<Float>& operator=(const MatrixGPU<Float>& m);
	Matrix<Float> operator*(const Matrix<Float>& m) const;
	void setUsrData(const QUIC::Uint& nr, const QUIC::Uint& nc, Float * d, const QUIC::Uint& p);
	Float * row(const QUIC::Uint& r);
	const Float * row(const QUIC::Uint& r) const;
	void realloc(const QUIC::Uint& nr, const QUIC::Uint& nc);
	void dealloc();
	virtual void copySubset(const QUIC::Uint * subsetRows, Matrix * A) const;
	bool svd(Matrix& U, Matrix& S, Matrix& VT) const;
	virtual Float norm() const;
	virtual void print(std::ostream& ost) const;
	template <typename FileType>
	void readFile(std::istream& ifs);
	static void mul(const Matrix& A, const Matrix& B, Matrix& C, bool AT, bool BT);
	QUIC::Uint nRows;
	QUIC::Uint nCols;
	bool usrData;
	QUIC::Uint pitch;
	Float * data;
protected:
	QUIC::MatAllocator<Float> * allocator;
	static QUIC::MatAllocator<Float> cpuAllocator;
};

template <typename Float>
class QUIC::MatrixGPU : public Matrix<Float>
{
public:
	MatrixGPU(const QUIC::Uint& nr, const QUIC::Uint& nc, Float * d, const QUIC::Uint& p);
	MatrixGPU(const QUIC::Uint& nr, const QUIC::Uint& nc);
	MatrixGPU(const Matrix<Float>& m);
	MatrixGPU();
	~MatrixGPU();
	MatrixGPU<Float>& operator=(const Matrix<Float>& m);
	MatrixGPU<Float>& operator=(const MatrixGPU<Float>& m);
	MatrixGPU<Float> operator*(const MatrixGPU<Float>& m) const;
	void copySubset(const QUIC::Uint * subsetRows, Matrix<Float> * A) const;
	bool svd(MatrixGPU& U, MatrixGPU& S, MatrixGPU& VT) const;
	void print(std::ostream& ost) const;
	static void mul(const MatrixGPU& A, const MatrixGPU& B, MatrixGPU& C, bool AT, bool BT);
protected:
	static MatAllocatorGPU<Float> gpuAllocator;
};

template <typename Float>
inline void QUIC::Matrix<Float>::setUsrData(const QUIC::Uint& nr, const QUIC::Uint& nc, Float * d, const QUIC::Uint& p)
{
	dealloc();
	nRows = nr;
	nCols = nc;
	data = d;
	pitch = p;
	usrData = true;
}

template <typename Float>
inline Float * QUIC::Matrix<Float>::row(const QUIC::Uint& r)
{
	return reinterpret_cast<Float *>(reinterpret_cast<char *>(data) + r*pitch);
}

template <typename Float>
inline const Float * QUIC::Matrix<Float>::row(const QUIC::Uint& r) const
{
	return reinterpret_cast<Float *>(reinterpret_cast<char *>(data) + r*pitch);
}

template <typename Float>
inline void QUIC::Matrix<Float>::dealloc()
{
	if (data) allocator->deallocate(MatAllocator<Float>::make_pitched_pointer(data, pitch), nCols, nRows);
	data = NULL;
	nRows = 0;
	nCols = 0;
}