#include "Partition.h"
#include <limits>
#include "../debugPrint.h"

//---------------------------------------------------------------------------------------------------------------

// Partition class

template <typename Float>
Partition<Float>::Partition(std::ifstream& ifs, const QUIC::Uint& r, const QUIC::Uint& nr, const QUIC::Uint& nc, const Float& e, const Float& d, const QUIC::OrthoMode& om, const QUIC::Uint& mr, const Partition * p)
: source(QUIC_SAFE_NEW(DiskPartitionSource<Float>, ifs, nr, nc)), nRows(nr), r0(r), nCols(nc), partition(NULL), cosineTree(NULL), orthoMode(om), Q(), rowSampler(NULL), 
  mcSqErr(-1.0), maxError(std::numeric_limits<Float>::infinity()), rootNSamples(0uLL), epsilon(e), delta(d), maxRank(mr)
{
	master = p ? p : this;
}

template <typename Float>
Partition<Float>::Partition(typename QUIC::MatAllocator<Float>::pitched_pointer& data, const QUIC::Uint& r, const QUIC::Uint& nr, const QUIC::Uint& nc, const Float& e, const Float& d, const QUIC::OrthoMode& om, const QUIC::Uint& mr, const Partition * p)
: source(QUIC_SAFE_NEW(RAMPartitionSource<Float>, data, nr, nc)), nRows(nr), r0(r), nCols(nc), partition(NULL), cosineTree(NULL), orthoMode(om), Q(), rowSampler(NULL), 
  mcSqErr(-1.0), maxError(std::numeric_limits<Float>::infinity()), rootNSamples(0uLL), epsilon(e), delta(d), maxRank(mr)
{
	master = p ? p : this;
}

template <typename Float>
Partition<Float>::~Partition()
{
	sleep();
	if (cosineTree) delete cosineTree;
	cosineTree = NULL;
}

template <typename Float>
void Partition<Float>::wake()
{
	if (partition == NULL)
	{
		partition = source->get();
		if (partition == NULL) return;
	}
	if (cosineTree == NULL) 
	{
		if (this == master) cosineTree = QUIC_SAFE_NEW(CosineTree, partition, orthoMode, maxRank, NULL);
		else cosineTree = QUIC_SAFE_NEW(CosineTree, partition, orthoMode, maxRank, master->cosineTree);
		Q.push(cosineTree->root, 0.0f);
	}
	cosineTree->A = partition;
	if (rowSampler == NULL)
	{
		rowSampler = QUIC_SAFE_NEW(LSDistribution<Float>::RowSampler, cosineTree->root);
	}
	if (mcSqErr < 0.0f)
	{
		mcSqErr = static_cast<Float>(cosineTree->nASq);
		maxError = epsilon*mcSqErr;
		rootNSamples = cosineTree->root->nSamples();
	}
}

template <typename Float>
void Partition<Float>::sleep()
{
	source->release();
	partition = NULL;
}

template <typename Float>
QUIC::Uint Partition<Float>::run()
{
	if (partition == NULL) wake();
	LSDistribution<Float>::RowSampler& rsRoot = *rowSampler;
	while (!Q.empty())
	{
		mcSqErr = cosineTree->MCSqError(rsRoot, delta, rootNSamples);
		DEBUG_PRINT("Pre-split error: %f\n", mcSqErr);
		if (mcSqErr <= maxError) break;
		CTNode<Float> * N = Q.pop();
		CTNode<Float> * C = N->split();
		if (C == NULL) continue;
		for (QUIC::Int i=0; i<2; ++i)
		{
			LSDistribution<Float>::RowSampler rs(C + i);
			Float e = cosineTree->MCSqError(rs, delta, C[i].nSamples());
			DEBUG_PRINT("Error so far = %f, %d rows in basis\n", e, cosineTree->commonVars.basis->size());
			Q.push(C + i, e);
		}
	}
	return cosineTree->commonVars.basis->size();
}

template <typename Float>
void Partition<Float>::project(QUIC::Matrix<Float>& basisMatrix, QUIC::Matrix<Float>& projection)
{
	if (partition == NULL) wake();
	QUIC::Matrix<Float> projectionSubMatrix(nRows, projection.nCols, projection.row(r0), projection.pitch);
	QUIC::Matrix<Float>::mul(*partition, basisMatrix, projectionSubMatrix, false, true);
}

//---------------------------------------------------------------------------------------------------------------

// PartitionGPU class

template <typename Float>
PartitionGPU<Float>::PartitionGPU(std::ifstream& ist, const QUIC::Uint& r, const QUIC::Uint& nr, const QUIC::Uint& nc, const Float& e, const Float& d, const QUIC::OrthoMode& om, const QUIC::Uint& mr, const PartitionGPU * p)
: Partition<Float>(ist, r, nr, nc, e, d, om, mr, p)
{}

template <typename Float>
PartitionGPU<Float>::PartitionGPU(typename QUIC::MatAllocator<Float>::pitched_pointer& data, const QUIC::Uint& r, const QUIC::Uint& nr, const QUIC::Uint& nc, const Float& e, const Float& d, const QUIC::OrthoMode& om, const QUIC::Uint& mr, const PartitionGPU * p)
: Partition<Float>(data, r, nr, nc, e, d, om, mr, p)
{}

template <typename Float>
void PartitionGPU<Float>::wake()
{
	if (partition == NULL)
	{
		QUIC::Matrix<Float> * temp = source->get();
		partition = QUIC_SAFE_NEW(QUIC::MatrixGPU<Float>, *temp);
		source->release();
	}
	if (cosineTree == NULL)
	{
		if (this == master) cosineTree = QUIC_SAFE_NEW(CosineTreeGPU<Float>, dynamic_cast<QUIC::MatrixGPU<Float> *>(partition), orthoMode, maxRank, NULL);
		else cosineTree = QUIC_SAFE_NEW(CosineTreeGPU<Float>, dynamic_cast<QUIC::MatrixGPU<Float> *>(partition), orthoMode, maxRank, dynamic_cast<CosineTreeGPU<Float> *>(master->cosineTree));
		Q.push(cosineTree->root, 0.0f);
	}
	Partition<Float>::wake();
}

template <typename Float>
void PartitionGPU<Float>::sleep()
{
	if (partition) delete partition;
	partition = NULL;
}

template <typename Float>
void PartitionGPU<Float>::project(QUIC::Matrix<Float>& basisMatrix, QUIC::Matrix<Float>& projection)
{
	if (partition == NULL) wake();
	QUIC::MatrixGPU<Float>& projectionGPU = dynamic_cast<QUIC::MatrixGPU<Float>&>(projection);
	QUIC::MatrixGPU<Float> * partitionGPU = dynamic_cast<QUIC::MatrixGPU<Float> *>(partition);
	QUIC::MatrixGPU<Float>& basisGPU = dynamic_cast<QUIC::MatrixGPU<Float>&>(basisMatrix);
	QUIC::MatrixGPU<Float> projectionSubMatrix(nRows, projectionGPU.nCols, projectionGPU.row(r0), projectionGPU.pitch);
	QUIC::MatrixGPU<Float>::mul(*partitionGPU, basisGPU, projectionSubMatrix, false, true);
}

template <typename Float>
QUIC::Matrix<Float> * PartitionGPU<Float>::fileBuffer = NULL;

INSTANTIATE_CLASS_TEMPLATE(Partition)