#pragma once

#include "QUIC.h"
#include <fstream>

template <typename Float>
class PartitionSource : public QUIC::Base
{
public:
	PartitionSource(const typename QUIC::MatAllocator<Float>::pitched_pointer& p, const QUIC::Uint& nr, const QUIC::Uint& nc);
	virtual ~PartitionSource();
	virtual QUIC::Matrix<Float> * get() = 0;
	virtual void release() = 0;
protected:
	QUIC::Matrix<Float> mat;
	QUIC::Uint nRows;
	QUIC::Uint nCols;
};

template <typename Float>
class DiskPartitionSource : public PartitionSource<Float>
{
public:
	DiskPartitionSource(std::ifstream& source, const QUIC::Uint& nr, const QUIC::Uint& nc);
	~DiskPartitionSource();
	virtual QUIC::Matrix<Float> * get();
	virtual void release();
private:
	std::ifstream& file;
	std::streampos filePos;
};

template <typename Float>
class RAMPartitionSource : public PartitionSource<Float>
{
public:
	RAMPartitionSource(typename QUIC::MatAllocator<Float>::pitched_pointer& source, const QUIC::Uint& nr, const QUIC::Uint& nc);
	virtual QUIC::Matrix<Float> * get();
	virtual void release();
private:
	typename QUIC::MatAllocator<Float>::pitched_pointer& data;
};

//---------------------------------------------------------------------------------------------------------------

// PartitionSource class

template <typename Float>
PartitionSource<Float>::PartitionSource(const typename QUIC::MatAllocator<Float>::pitched_pointer& p, const QUIC::Uint& nr, const QUIC::Uint& nc)
: mat(nr, nc, p.addr, p.pitch), nRows(nr), nCols(nc)
{}

template <typename Float>
PartitionSource<Float>::~PartitionSource()
{}

//---------------------------------------------------------------------------------------------------------------

// DiskPartitionSource class

template <typename Float>
DiskPartitionSource<Float>::DiskPartitionSource(std::ifstream& source, const QUIC::Uint& nr, const QUIC::Uint& nc)
: PartitionSource(QUIC::MatAllocator<Float>::make_pitched_pointer(NULL, 0), nr, nc), file(source), filePos(source.tellg())
{
	source.seekg(nRows*nCols*sizeof(Float), std::ios_base::cur);
	if (source.fail()) source.clear();
}

template <typename Float>
DiskPartitionSource<Float>::~DiskPartitionSource()
{
	release();
}

template <typename Float>
QUIC::Matrix<Float> * DiskPartitionSource<Float>::get()
{
	if (mat.data == NULL) mat.realloc(nRows, nCols);
	file.seekg(filePos);
	bool fileError = file.fail();
	if (!fileError) 
	{
		mat.readFile<Float>(file);
		fileError = file.fail() || file.eof();
	}
	if (fileError)
	{
		release();
		throw (QUICException(__FILE__, __LINE__, "Failed to read file."));
	}
	return &mat;
};

template <typename Float>
void DiskPartitionSource<Float>::release()
{
	mat.dealloc();
}

//---------------------------------------------------------------------------------------------------------------

// RAMPartitionSource class

template <typename Float>
RAMPartitionSource<Float>::RAMPartitionSource(typename QUIC::MatAllocator<Float>::pitched_pointer& source, const QUIC::Uint& nr, const QUIC::Uint& nc)
: PartitionSource(source, nr, nc), data(source)
{
	source.addr = arrayRow(source.addr, nr, source.pitch);
}

template <typename Float>
QUIC::Matrix<Float> * RAMPartitionSource<Float>::get()
{
	return &mat;
}

template <typename Float>
void RAMPartitionSource<Float>::release()
{}
