#include "PQueue.h"
#include "globalFuncs.h"
#include "CosineTree.h"
#include "Matrix.h"
#include "profile.h"
#include "Timer.h"
#include "omp.h"
#include "Partition.h"
#include <limits>
#include "QUICException.h"
#include "../debugPrint.h"

static bool initialized = false;
static bool cpu = false;
static bool gpu = false;
static QUIC::Int gpuArch;
size_t maxPitch;

std::string errorMsg;
bool error = false;

template <typename Float>
bool checkFormat(std::ifstream& ifs)
{
	std::string format;
	//getline(ifs, format, '\0');
	getline(ifs, format); // rui

	/*const size_t L = strlen("float");
	bool result = format.size() > 5 && format.substr(0, L) == "float";
	if (result)
	{
		size_t floatSize = static_cast<size_t>(atoi(format.c_str() + L));
		if (floatSize != 8uLL*sizeof(Float)) result = false;
	}
	if (!result)
	{
		setError("Invalid data type");
	}
	return result;*/
	return true; // rui
}

// non-partitioned

#ifdef PROFILE
template <typename CTType, typename Float, TimerType tt>
#else
template <typename CTType, typename Float>
#endif
inline QUIC::Uint runQUIC(const typename CTType::Matrix * A, typename CTType::Matrix * U, typename CTType::Matrix * S, typename CTType::Matrix * VT,
						  Float epsilon, Float delta, const QUIC::OrthoMode& orthoMode, const QUIC::Uint& maxRank)
{
	QUIC::Uint rval = 0;
	try
	{
		START_TIMER(BUILDING_TREE, tt);
		PQueue<CTNode<Float> *, Float> Q;
		CTType cosineTree(A, orthoMode, maxRank, NULL);
		Q.push(cosineTree.root, 0.0f);
		LSDistribution<Float>::RowSampler rs(cosineTree.root);
		Float mcSqErr = static_cast<Float>(cosineTree.nASq);
		Float maxError = epsilon*mcSqErr;
		QUIC::Uint rootNSamples = cosineTree.root->nSamples();
		Basis<Float>& basis = *cosineTree.commonVars.basis;
		while (!Q.empty() && mcSqErr > maxError && (maxRank == 0 || basis.size() < maxRank))
		{
			CTNode<Float> * N = Q.pop();
			CTNode<Float> * C = N->split();
			if (C == NULL) continue;
			for (QUIC::Int i=0; i<2; ++i)
			{
				LSDistribution<Float>::RowSampler rs(C + i);
				Float e = cosineTree.MCSqError(rs, delta, C[i].nSamples());
				Q.push(C + i, e);
			}
			mcSqErr = cosineTree.MCSqError(rs, delta, rootNSamples);
		}
		STOP_TIMER(BUILDING_TREE, tt);
		rval = cosineTree.commonVars.basis->size();
		START_TIMER(EXTRACTING_SVD, tt);
		cosineTree.extractSVD(U, S, VT);
		STOP_TIMER(EXTRACTING_SVD, tt);
	}
	catch (QUICException e)
	{
		setError(e.msg);
		rval = 0;
	}
	return rval;
}

// partitioned

#ifdef PROFILE
template <typename PartitionType, typename InputType, typename Float, TimerType tt>
#else
template <typename PartitionType, typename InputType, typename Float>
#endif
inline QUIC::Uint runQUIC(InputType& input, const QUIC::Uint& nRows, const QUIC::Uint& nCols, typename PartitionType::CosineTree::Matrix * U, 
						  typename PartitionType::CosineTree::Matrix * S, typename PartitionType::CosineTree::Matrix * VT, Float epsilon, Float delta,
						  const QUIC::OrthoMode& orthoMode, const QUIC::Uint nBlocks, const QUIC::Uint& maxRank)
{
	QUIC::Uint rval = 0;
	QUIC::Uint partitionNRows = iDivUp(nRows, nBlocks);
	QUIC::Uint partitionSize = nCols*partitionNRows;
	PQueue<PartitionType *, Float> Q;
	PartitionType ** partitions = NULL;
	QUIC::Uint partitionStart = 0uLL;
	DEBUG_PRINT("%d rows per partition\n", partitionNRows);
	try
	{
		if (nBlocks == 0 || nBlocks > nRows) throw(QUICException(__FILE__, __LINE__, "Invalid number of partitions"));
		START_TIMER(BUILDING_TREE, tt);
		partitions = QUIC_SAFE_ALLOC(PartitionType *, nBlocks);
		for (QUIC::Uint i=0; i<nBlocks; ++i) partitions[i] = NULL;
		Basis<Float> * basis = NULL;
		for (QUIC::Uint i=0; i<nBlocks; ++i)
		{
			PartitionType *& partition = partitions[i];
			QUIC::Uint partitionSize = partitionNRows;
			QUIC::Uint partitionEnd = std::min(partitionStart + partitionNRows, nRows);
			DEBUG_PRINT("Creating partition %d (rows %d - %d)\n", i, partitionStart, partitionEnd);
			partition = QUIC_SAFE_NEW(PartitionType, input, partitionStart, partitionEnd - partitionStart, nCols, epsilon, delta, orthoMode, maxRank, partitions[0]);
			partitionStart += partitionNRows;
			DEBUG_PRINT("Running partition\n");
			partition->run();
			DEBUG_PRINT("Partition error: %f\n", partition->mcSqErr);
			DEBUG_PRINT("Deactivating partition\n");
			partition->sleep();
			if (!partition->Q.empty())
			{
				DEBUG_PRINT("Placing partition in queue\n");
				Q.push(partition, partition->mcSqErr);
			}
			basis = partitions[0]->cosineTree->commonVars.basis;
			//std::cout<<"Push "<<partition->r0<<", "<<partition->mcSqErr<<std::endl;
			//std::cout<<(partition->cosineTree->commonVars.basis->size())<<" basis vector so far"<<std::endl;
		}
		DEBUG_PRINT("Initial pass complete\n");
		CosineTree<Float>::CommonVars& cv = partitions[0]->cosineTree->commonVars;
		double nASq = 0.0f;
		Float mcSqErr = 0.0f;
		for (int i=0; i<nBlocks; ++i)
		{
			nASq += partitions[i]->cosineTree->nASq;
			mcSqErr += partitions[i]->mcSqErr;
		}
		Float maxError = static_cast<Float>(static_cast<double>(epsilon)*nASq);
		DEBUG_PRINT("Error: %f\n", mcSqErr);
		while (!Q.empty() && mcSqErr > maxError && (maxRank == 0 || basis->size() < maxRank))
		{
			Float k = Q.nextKey();
			PartitionType * P = Q.pop();
			DEBUG_PRINT("Reactivating partition %d\n", P->r0/partitionNRows);
			//std::cout<<"Pop "<<P->r0<<", "<<k<<std::endl;
			mcSqErr -= P->mcSqErr;
			P->maxError -= (mcSqErr - maxError)/static_cast<Float>(nBlocks);
			DEBUG_PRINT("Running partition\n");
			P->run();
			mcSqErr += P->mcSqErr;
			DEBUG_PRINT("Deactivating partition\n");
			P->sleep();
			//std::cout<<"Push "<<P->r0<<", "<<P->mcSqErr<<std::endl;
			//std::cout<<(P->cosineTree->commonVars.basis->size())<<" basis vector so far"<<std::endl;
			if (!P->Q.empty())
			{
				Q.push(P, P->mcSqErr);
				DEBUG_PRINT("Placing partition in queue\n");
			}
		}
		DEBUG_PRINT("Second pass complete\n");
		rval = basis->size();
		DEBUG_PRINT("Basis size: %d\n", rval);
		PartitionType::CosineTree::Matrix VhatT;
		DEBUG_PRINT("Extracting basis\n");
		basis->getMatrix(VhatT);
		PartitionType::CosineTree::Matrix AVhat(nRows, basis->size());
		for (int i=0; i<nBlocks; ++i)
		{
			DEBUG_PRINT("Projecting partition %d\n", i);
			//std::cout<<"Projecting partition "<<i<<std::endl;
			partitions[i]->project(VhatT, AVhat);
			partitions[i]->sleep();
		}
		STOP_TIMER(BUILDING_TREE, tt);
		//std::cout<<"Extracting SVD"<<std::endl;
		DEBUG_PRINT("Extracting svd\n");
		START_TIMER(EXTRACTING_SVD, tt);
		extractSVDFromProjection<PartitionType::CosineTree::Matrix, Float>(AVhat, VhatT, U, S, VT, cv.scratchRow<Float>(0), cv.scratchPitch);
		STOP_TIMER(EXTRACTING_SVD, tt);
	}
	catch (QUICException e)
	{
		setError(e.msg);
		rval = 0;
	}
	DEBUG_PRINT("Freeing partitions\n");
	for (int i=0; i<nBlocks; ++i) 
	{
		if (partitions[i]) delete partitions[i];
	}
	hostFree(partitions);
	return rval;
}

// Non-partitioned matrix input, CPU

template <typename Float>
QUIC::Uint QUIC::QUIC_SVD(const Matrix<Float> * A, Matrix<Float> * U, Matrix<Float> * S, Matrix<Float> * VT, const Float& epsilon, 
						  const Float& delta, const OrthoMode& orthoMode, const QUIC::Uint& maxRank)
{
#ifdef PROFILE
	QUIC::Uint rval = runQUIC<CosineTree<Float>, Float, CPU>(A, U, S, VT, epsilon, delta, orthoMode, maxRank);
#else
	QUIC::Uint rval = runQUIC<CosineTree<Float>, Float>(A, U, S, VT, epsilon, delta, orthoMode, maxRank);
#endif
	if (error) rval = 0;
	else if (rval) cpu = true;
	return rval;
}

// Non-partitioned matrix input, GPU

template <typename Float>
QUIC::Uint QUIC::QUIC_SVD(const MatrixGPU<Float> * A, MatrixGPU<Float> * U, MatrixGPU<Float> * S, MatrixGPU<Float> * VT, 
						  const Float& epsilon, const Float& delta, const QUIC::OrthoMode& orthoMode, const QUIC::Uint& maxRank)
{
	if (!initialized)
	{
		errorMsg = "GPU is not initialized.";
		error = true;
		return 0;
	}
	if (!checkArch<Float>(gpuArch)) return 0;
#ifdef PROFILE
	QUIC::Uint rval = runQUIC<CosineTreeGPU<Float>, Float, GPU>(A, U, S, VT, epsilon, delta, orthoMode, maxRank);
#else
	QUIC::Uint rval = runQUIC<CosineTreeGPU<Float>, Float>(A, U, S, VT, epsilon, delta, orthoMode, maxRank);
#endif
	if (error) rval = 0;
	else if (rval) gpu = true;
	return rval;
}

// Partitioned file input, CPU

template <typename Float>
QUIC::Uint QUIC::QUIC_SVD(std::ifstream& ist, QUIC::Matrix<Float> * U, QUIC::Matrix<Float> * S, QUIC::Matrix<Float> * VT, 
						  const Float& epsilon, const Float& delta, const QUIC::OrthoMode& orthoMode, const QUIC::Uint nBlocks, const QUIC::Uint& maxRank)
{
	if (!checkFormat<Float>(ist)) return false;
	int size [2];
	ist.read(reinterpret_cast<char *>(size), 2*sizeof(int));
	QUIC::Uint nRows = static_cast<QUIC::Uint>(size[0]);
	QUIC::Uint nCols = static_cast<QUIC::Uint>(size[1]);
	if (nBlocks == 1)
	{
		Matrix<Float> A(ist);
		if (checkError()) return 0;
		return QUIC_SVD<Float>(&A, U, S, VT, epsilon, delta, orthoMode, maxRank);
	}
	else
	{
#ifdef PROFILE
		QUIC::Uint rval = runQUIC<Partition<Float>, std::ifstream, Float, CPU>(ist, nRows, nCols, U, S, VT, epsilon, delta, orthoMode, nBlocks, maxRank);
#else
		QUIC::Uint rval = runQUIC<Partition<Float>, std::ifstream, Float>(ist, nRows, nCols, U, S, VT, epsilon, delta, orthoMode, nBlocks, maxRank);
#endif
		if (error) rval = 0;
		else if (rval) cpu = true;
		return rval;
	}
	return 0;
}

// Partitioned file input, GPU

template <typename Float>
QUIC::Uint QUIC::QUIC_SVD(std::ifstream& ist, QUIC::MatrixGPU<Float> * U, QUIC::MatrixGPU<Float> * S, QUIC::MatrixGPU<Float> * VT, 
						  const Float& epsilon, const Float& delta, const QUIC::OrthoMode& orthoMode, const QUIC::Uint nBlocks, const QUIC::Uint& maxRank)
{
	if (!checkFormat<Float>(ist)) return false;
	int size [2];
	//ist.read(reinterpret_cast<char *>(size), 2*sizeof(int));
	//QUIC::Uint nRows = static_cast<QUIC::Uint>(size[0]);
	//QUIC::Uint nCols = static_cast<QUIC::Uint>(size[1]);
	std::string line;
	getline(ist, line);
	sscanf(line.c_str(), "%d", &size[0]);
	getline(ist, line);
	sscanf(line.c_str(), "%d", &size[1]);

	QUIC::Uint nRows = size[0];
	QUIC::Uint nCols = size[1];

	getline(ist, line);
	printf("Matrix dimension: %d %d\n", (int)nRows, (int)nCols);
	printf("last line: %s\n", line.c_str());
	if (nBlocks == 1)
	{
		Matrix<Float> * temp = QUIC_SAFE_NEW(QUIC::Matrix<Float>, ist);
		if (checkError()) return 0;
		MatrixGPU<Float> A(*temp);
		delete temp;
		if (checkError()) return 0;
		return QUIC_SVD<Float>(&A, U, S, VT, epsilon, delta, orthoMode,	maxRank);
	}
	else
	{
		if (!initialized)
		{
			errorMsg = "GPU is not initialized.";
			error = true;
			return 0;
		}
		if (!checkArch<Float>(gpuArch)) return 0;
#ifdef PROFILE
		QUIC::Uint rval = runQUIC<PartitionGPU<Float>, std::ifstream, Float, GPU>(ist, nRows, nCols, U, S, VT, epsilon, delta, orthoMode, nBlocks, maxRank);
#else
		QUIC::Uint rval = runQUIC<PartitionGPU<Float>, std::ifstream, Float>(ist, nRows, nCols, U, S, VT, epsilon, delta, orthoMode, nBlocks, maxRank);
#endif
		QUIC::Matrix<Float> hostS; 
		hostS = *S;

		for (int s=0; s<(int) nCols; s++) {
			printf("%f\n", hostS.data[s]);
		}

		if (error) rval = 0;
		else if (rval) gpu = true;
		return rval;
	}
}

// Partitioned matrix input, CPU

template <typename Float>
QUIC::Uint QUIC::QUIC_SVD(const QUIC::Matrix<Float> * A, QUIC::Matrix<Float> * U, QUIC::Matrix<Float> * S, QUIC::Matrix<Float> * VT, 
						  const Float& epsilon, const Float& delta, const QUIC::OrthoMode& orthoMode, const QUIC::Uint nBlocks, const QUIC::Uint& maxRank)
{
	if (nBlocks == 1)
	{
		return QUIC_SVD<Float>(A, U, S, VT, epsilon, delta, orthoMode, maxRank);
	}
	else
	{
		MatAllocator<Float>::pitched_pointer p = MatAllocator<Float>::make_pitched_pointer(A->data, A->pitch);
#ifdef PROFILE
		QUIC::Uint rval = runQUIC<Partition<Float>, MatAllocator<Float>::pitched_pointer, Float, CPU>(p, A->nRows, A->nCols, U, S, VT, epsilon, delta, orthoMode, nBlocks, maxRank);
#else
		QUIC::Uint rval = runQUIC<Partition<Float>, MatAllocator<Float>::pitched_pointer, Float>(p, A->nRows, A->nCols, U, S, VT, epsilon, delta, orthoMode, nBlocks, maxRank);
#endif
		if (error) rval = 0;
		else if (rval) cpu = true;
		return rval;
	}
	return 0;
}

// Partitioned matrix input, GPU

template <typename Float>
QUIC::Uint QUIC::QUIC_SVD(const QUIC::Matrix<Float> * A, QUIC::MatrixGPU<Float> * U, QUIC::MatrixGPU<Float> * S, QUIC::MatrixGPU<Float> * VT, 
						  const Float& epsilon, const Float& delta, const QUIC::OrthoMode& orthoMode, const QUIC::Uint nBlocks, const QUIC::Uint& maxRank)
{
	if (nBlocks == 1)
	{
		DEBUG_PRINT("Copying input to GPU");
		MatrixGPU<Float> Agpu(*A);
		DEBUG_PRINT("Running QUIC");
		return QUIC_SVD<Float>(&Agpu, U, S, VT, epsilon, delta, orthoMode, maxRank);		
	}
	else
	{
		if (!initialized)
		{
			errorMsg = "GPU is not initialized.";
			error = true;
			return 0;
		}
		if (!checkArch<Float>(gpuArch)) return 0;
		MatAllocator<Float>::pitched_pointer p = MatAllocator<Float>::make_pitched_pointer(A->data, A->pitch);
		DEBUG_PRINT("Running QUIC\n");
#ifdef PROFILE
		QUIC::Uint rval = runQUIC<PartitionGPU<Float>, MatAllocator<Float>::pitched_pointer, Float, GPU>(p, A->nRows, A->nCols, U, S, VT, epsilon, delta, orthoMode, nBlocks, maxRank);
#else
		QUIC::Uint rval = runQUIC<PartitionGPU<Float>, MatAllocator<Float>::pitched_pointer, Float>(p, A->nRows, A->nCols, U, S, VT, epsilon, delta, orthoMode, nBlocks, maxRank);
#endif
		if (error) rval = 0;
		else if (rval) gpu = true;
		return rval;
	}
}

bool QUIC::initGPU()
{
	if (initialized) return true;
	int deviceCount = 0;
	cudaGetDeviceCount(&deviceCount);
	if (!deviceCount)
	{
		setError("No CUDA enabled device available.");
		return false;
	}
	int device = cutGetMaxGflopsDeviceId();
	cudaSetDevice(device);
	cudaDeviceProp properties;
	if (cudaGetDeviceProperties(&properties, device) != cudaSuccess)
	{
		setError("No CUDA enabled device available.");
		return false;
	}
	gpuArch = 10*properties.major + properties.minor;
	maxPitch = properties.memPitch;
	cublasInit();
	culaInitialize();
	initialized = true;
	return true;
}

void QUIC::shutdownGPU()
{

	culaShutdown();
	cublasShutdown();
	cudaThreadExit();
	initialized = false;
}

void QUIC::printTimers(std::ostream& ost)
{
#ifdef PROFILE
	QUIC::Uint t = 0;
	if (cpu) t |= (1 << CPU);
	if (gpu) t |= (1 << GPU);
	::PRINT_TIMERS(t, ost);
#else
	ost<<"Profiling is not supported."<<std::endl;
#endif
}

void QUIC::resetTimers()
{
	cpu = false;
	gpu = false;
	RESET_TIMERS();
}

bool QUIC::checkError()
{
	return error;
}

void QUIC::setBlockSize(const QUIC::Uint& s)
{
	Basis<float>::mgsBlockSize = s;
	Basis<double>::mgsBlockSize = s;
}

const std::string * QUIC::getLastError()
{
	if (error)
	{	
		error = false;
		return &errorMsg;
	}
	return NULL;
}

template QUIC::Uint QUIC::QUIC_SVD(const Matrix<float> * A, Matrix<float> * U, Matrix<float> * S, Matrix<float> * VT, const float& epsilon, const float& delta, const OrthoMode& orthoMode, const QUIC::Uint& maxRank);
template QUIC::Uint QUIC::QUIC_SVD(const Matrix<double> * A, Matrix<double> * U, Matrix<double> * S, Matrix<double> * VT, const double& epsilon, const double& delta, const OrthoMode& orthoMode, const QUIC::Uint& maxRank);
template QUIC::Uint QUIC::QUIC_SVD(const MatrixGPU<float> * A, MatrixGPU<float> * U, MatrixGPU<float> * S, MatrixGPU<float> * VT, const float& epsilon, const float& delta, const OrthoMode& orthoMode, const QUIC::Uint& maxRank);
template QUIC::Uint QUIC::QUIC_SVD(const MatrixGPU<double> * A, MatrixGPU<double> * U, MatrixGPU<double> * S, MatrixGPU<double> * VT, const double& epsilon, const double& delta, const OrthoMode& orthoMode, const QUIC::Uint& maxRank);
template QUIC::Uint QUIC::QUIC_SVD(std::ifstream& ist, QUIC::Matrix<float> * U, QUIC::Matrix<float> * S, QUIC::Matrix<float> * VT, const float& epsilon, const float& delta, const QUIC::OrthoMode& orthoMode, const QUIC::Uint nBlocks, const QUIC::Uint& maxRank);
template QUIC::Uint QUIC::QUIC_SVD(std::ifstream& ist, QUIC::Matrix<double> * U, QUIC::Matrix<double> * S, QUIC::Matrix<double> * VT, const double& epsilon, const double& delta, const QUIC::OrthoMode& orthoMode, const QUIC::Uint nBlocks, const QUIC::Uint& maxRank);
template QUIC::Uint QUIC::QUIC_SVD(std::ifstream& ist, QUIC::MatrixGPU<float> * U, QUIC::MatrixGPU<float> * S, QUIC::MatrixGPU<float> * VT, const float& epsilon, const float& delta, const QUIC::OrthoMode& orthoMode, const QUIC::Uint nBlocks, const QUIC::Uint& maxRank);
template QUIC::Uint QUIC::QUIC_SVD(std::ifstream& ist, QUIC::MatrixGPU<double> * U, QUIC::MatrixGPU<double> * S, QUIC::MatrixGPU<double> * VT, const double& epsilon, const double& delta, const QUIC::OrthoMode& orthoMode, const QUIC::Uint nBlocks, const QUIC::Uint& maxRank);
template QUIC::Uint QUIC::QUIC_SVD(const Matrix<float> * A, QUIC::Matrix<float> * U, QUIC::Matrix<float> * S, QUIC::Matrix<float> * VT, const float& epsilon, const float& delta, const QUIC::OrthoMode& orthoMode, const QUIC::Uint nBlocks, const QUIC::Uint& maxRank);
template QUIC::Uint QUIC::QUIC_SVD(const Matrix<double> * A, QUIC::Matrix<double> * U, QUIC::Matrix<double> * S, QUIC::Matrix<double> * VT, const double& epsilon, const double& delta, const QUIC::OrthoMode& orthoMode, const QUIC::Uint nBlocks, const QUIC::Uint& maxRank);
template QUIC::Uint QUIC::QUIC_SVD(const Matrix<float> * A, QUIC::MatrixGPU<float> * U, QUIC::MatrixGPU<float> * S, QUIC::MatrixGPU<float> * VT, const float& epsilon, const float& delta, const QUIC::OrthoMode& orthoMode, const QUIC::Uint nBlocks, const QUIC::Uint& maxRank);
template QUIC::Uint QUIC::QUIC_SVD(const Matrix<double> * A, QUIC::MatrixGPU<double> * U, QUIC::MatrixGPU<double> * S, QUIC::MatrixGPU<double> * VT, const double& epsilon, const double& delta, const QUIC::OrthoMode& orthoMode, const QUIC::Uint nBlocks, const QUIC::Uint& maxRank);