#include "globalFuncs.h"
#include <map>
#include <cassert>
#include "Matrix.h"
#include "profile.h"

std::map<void *, QUIC::Uint> mem;
QUIC::Uint allocated = 0;
QUIC::Uint maxAllocated = 0;
QUIC::Uint failures = 0;

void devAllocTrack(void * p, QUIC::Uint n, char * file, int line)
{
	/*if (p)
	{
		mem[p] = n;
		allocated += n;
		if (allocated > maxAllocated) maxAllocated = allocated;
		std::cout<<"Allocated "<<(n/(1024*1024))<<" MB (total allocated = "<<(allocated/(1024*1024))<<" MB, max allocated = "
			<<(maxAllocated/(1024*1024))<<" MB, "<<mem.size()<<" chunks not freed, "<<failures<<" failures)"<<std::endl;
	}
	else
	{
		std::cout<<file<<", line "<<line<<": Allocation failed (tried to allocate "<<(n/(1024*1024))<<" MB, "<<allocated<<" MB already allocated)"<<std::endl;
	}*/
}

void devFreeTrack(void * p, char * file, int line)
{
	/*std::map<void *, QUIC::Uint>::iterator iter = mem.find(p);
	if (iter == mem.end()) 
	{
		std::cout<<file<<", line "<<line<<": Memory was never allocated!\a"<<std::endl;
	}
	else if (p == NULL)
	{
		std::cout<<file<<", line "<<line<<": NULL pointer!\a"<<std::endl;
	}
	else
	{
		allocated -= iter->second;
		std::cout<<"Freed "<<(iter->second/(1024*1024))<<" MB (total allocated = "<<(allocated/(1024*1024))<<" MB, max allocated = "
			<<(maxAllocated/(1024*1024))<<" MB, "<<mem.size()<<" chunks not freed, "<<failures<<" failures)"<<std::endl;
		mem.erase(p);
	}*/
}

template <typename Float>
void matrixSqrt(QUIC::Matrix<Float>& m)
{
	#pragma omp parallel for
	for (QUIC::Int r=0; r<m.nRows; ++r)
	{
		Float * row = m.row(r);
		for (QUIC::Uint c=0; c<m.nCols; ++c)
		{
			Float& elt = row[c];
			elt = Sqrt(elt);
		}
	}
}

template <typename Float>
void diagDivR(const QUIC::Matrix<Float>& A, const QUIC::Matrix<Float>& v, QUIC::Matrix<Float>& B)
{
	assert(A.nCols == v.nCols);
	assert(A.nRows == B.nRows);
	assert(A.nCols == B.nCols);
	const Float * vRow = v.row(0);
	QUIC::Uint ldA = A.pitch/sizeof(Float);
	QUIC::Uint ldB = B.pitch/sizeof(Float);
	for (QUIC::Uint c=0; c<A.nCols; ++c)
	{
		const Float * Aentry = A.row(0) + c;
		Float * Bentry = B.row(0) + c;
		Float s = 1.0f/vRow[c];
		for (QUIC::Uint r=0; r<A.nRows; ++r)
		{
			*Bentry = (*Aentry)*s;
			Aentry += ldA;
			Bentry += ldB;
		}
	}
}

template <typename Float>
void diagDivL(const QUIC::Matrix<Float>& A, const QUIC::Matrix<Float>& v, QUIC::Matrix<Float>& B)
{
	assert(A.nRows == v.nCols);
	const Float * vRow = v.row(0);
	for (QUIC::Uint r=0; r<A.nRows; ++r)
	{
		const Float * Arow = A.row(r);
		Float * Brow = B.row(r);
		if (&B != &A) cpuMemcpy<Float>(A.nCols, Arow, Brow);
		cpuScalarMul(A.nCols, 1.0f/vRow[r], Brow);
	}
}

template <typename Float>
void vectorMul(const QUIC::Matrix<Float>& x, QUIC::Matrix<Float>& y)
{
	assert(x.nRows == 1 && y.nRows == 1 && x.nCols == y.nCols);
	#pragma omp parallel for
	for (QUIC::Int i=0; i<x.nCols; ++i) y.data[i] *= x.data[i];
}

template <typename Float>
void rowNorms(const QUIC::Matrix<Float>& A, const bool& square, Float * result)
{
	#pragma omp parallel for
	for (QUIC::Int i=0; i<A.nRows; ++i)
	{
		const Float * row = A.row(i);
		Float norm = cpuNorm(A.nCols, row);
		result[i] = square ? norm*norm : norm;
	}
}

template <typename Float>
QUIC::Uint cpuLabelRows(Float * cosines, const QUIC::Uint& nRows, const Float& cMin, const Float& cMax, QUIC::Uint * labels, QUIC::Uint& blockSize, QUIC::Uint& nBlocks)
{
	blockSize = 8*sizeof(QUIC::Uint);
	nBlocks = iDivUp(nRows, blockSize);
	#pragma omp parallel for
	for (QUIC::Int i=0; i<nBlocks; ++i)
	{
		QUIC::Uint& blockLabels = labels[i];
		blockLabels = 0;
		QUIC::Uint blockStart = i*blockSize;
		QUIC::Uint blockN = std::min(blockSize, nRows - blockStart);
		Float * blockCosines = cosines + blockStart;
		for (QUIC::Uint j=0; j<blockN; ++j)
		{
			Float& cosine = blockCosines[j];
			if (cMax - cosine <= cosine - cMin) blockLabels |= (static_cast<QUIC::Uint>(1) << j);
		}
	}
	return nBlocks;
}

#ifdef PROFILE
template <typename Matrix>
TimerType timerType();

template <>
TimerType timerType<QUIC::Matrix<float> >()
{
	return CPU;
}

template <>
TimerType timerType<QUIC::Matrix<double> >()
{
	return CPU;
}

template <>
TimerType timerType<QUIC::MatrixGPU<float> >()
{
	return GPU;
}

template <>
TimerType timerType<QUIC::MatrixGPU<double> >()
{
	return GPU;
}
#endif

template <typename Matrix, typename Float>
bool extractSVDFromProjection(const Matrix& AVhat, const Matrix& VhatT, Matrix * U, Matrix * S, Matrix * VT, Float * scratch, const size_t& scratchPitch)
{
#ifdef PROFILE
	TimerType tt = timerType<Matrix>();
#endif
	Matrix& Sr = *S;
	Matrix AVhatTAVhat(AVhat.nCols, AVhat.nCols);
	Matrix::mul(AVhat, AVhat, AVhatTAVhat, true, false);
	Matrix Uprime, VprimeT;
	START_TIMER(SVD, tt);
	AVhatTAVhat.svd(Uprime, Sr, VprimeT);
	STOP_TIMER(SVD, tt);
	if (VT)
	{
		Matrix& VTr = *VT;
		VTr.realloc(VprimeT.nRows, VhatT.nCols);
		Matrix::mul(VprimeT, VhatT, VTr, false, false);
	}
	START_TIMER(SQRT, tt);
	matrixSqrt(Sr);
	STOP_TIMER(SQRT, tt);
	if (U)
	{
		Matrix& Ur = *U;
		Ur.realloc(AVhat.nRows, VprimeT.nRows);
		Matrix::mul(AVhat, VprimeT, Ur, false, true);
		START_TIMER(DIAG_DIV, tt);
		diagDivR(Ur, Sr, Ur);
		STOP_TIMER(DIAG_DIV, tt);
	}
	if (VT)
	{
		Matrix& VTr = *VT;
		Float * s = Sr.data;
		Matrix normMat(1, VTr.nRows, scratch, scratchPitch);
		rowNorms(VTr, false, normMat.data);
		diagDivL(VTr, normMat, VTr);
		vectorMul(normMat, Sr);
	}
	return true;
}

#include "explicit.h"