#include "globalFuncs.h"
#include <math_functions.h>
#include <cassert>
#include <iostream>
#include "Matrix.h"
#include "profile.h"
#include "kernels.h"

inline __host__ __device__ float Rsqrt(const float& x)
{
	return rsqrtf(x);
}

inline __host__ __device__ double Rsqrt(const double& x)
{
	return rsqrt(x);
}

template <typename Float>
void matrixSqrt(QUIC::MatrixGPU<Float>& m)
{
	const QUIC::Uint blockSize = 8;
	dim3 dimGrid(static_cast<unsigned int>(iDivUp(m.nCols, blockSize)), static_cast<unsigned int>(iDivUp(m.nRows, blockSize)));
	dim3 dimBlock(blockSize, blockSize);
	vectorSqrtKernel<Float><<<dimGrid, dimBlock>>>(m.data, m.pitch, m.nRows, m.nCols);
	QUIC_SAFE_CALL(cudaThreadSynchronize());
}

template <typename Float>
void diagDivR(const QUIC::MatrixGPU<Float>& A, const QUIC::MatrixGPU<Float>& v, QUIC::MatrixGPU<Float>& B)
{
	const QUIC::Uint blockSizeX = 16;
	const QUIC::Uint blockSizeY = 16;
	assert(v.nCols == A.nCols);
	dim3 dimGrid(static_cast<unsigned int>(iDivUp(A.nCols, blockSizeX)), static_cast<unsigned int>(iDivUp(A.nRows, blockSizeY)));
	dim3 dimBlock(blockSizeX, blockSizeY);
	diagDivRKernel<Float, blockSizeX><<<dimGrid, dimBlock>>>(A.data, A.pitch, v.data, B.data, B.pitch, A.nRows, A.nCols);
	QUIC_SAFE_CALL(cudaThreadSynchronize());
}

template <typename Float>
void diagDivL(const QUIC::MatrixGPU<Float>& A, const QUIC::MatrixGPU<Float>& v, QUIC::MatrixGPU<Float>& B)
{
	const QUIC::Uint blockSizeX = 16;
	const QUIC::Uint blockSizeY = 16;
	dim3 dimGrid(static_cast<unsigned int>(iDivUp(A.nCols, blockSizeX)), static_cast<unsigned int>(iDivUp(A.nRows, blockSizeY)));
	dim3 dimBlock(static_cast<unsigned int>(blockSizeX), static_cast<unsigned int>(blockSizeY));
	diagDivLKernel<Float, blockSizeX><<<dimGrid, dimBlock>>>(A.data, A.pitch, v.data, B.data, B.pitch, A.nRows, A.nCols);
	QUIC_SAFE_CALL(cudaThreadSynchronize());
}

template <typename Float>
void vectorMul(const QUIC::MatrixGPU<Float>& x, QUIC::MatrixGPU<Float>& y)
{
	const QUIC::Uint blockSize = 128;
	const QUIC::Uint& n = x.nCols;
	assert(x.nRows == 1 && y.nRows == 1 && x.nCols == y.nCols);
	dim3 dimGrid(static_cast<unsigned int>(iDivUp(n, blockSize)));
	dim3 dimBlock(static_cast<unsigned int>(blockSize));
	vectorMulKernel<Float><<<dimGrid, dimBlock>>>(x.data, y.data, n);
	QUIC_SAFE_CALL(cudaThreadSynchronize());
}

template <typename Float>
void rowNorms(const QUIC::MatrixGPU<Float>& A, const bool& square, Float * result)
{
	QUIC::Uint blockSize = 512;
	if (A.nCols < blockSize)
	{
		blockSize = 1uLL << log2(A.nCols);
		if (blockSize < A.nCols) blockSize <<= 1uLL;
	}
	dim3 dimBlock(static_cast<unsigned int>(blockSize), 1);
	dim3 dimGrid(static_cast<unsigned int>(A.nRows), 1);
	sqRowNormsKernel<Float><<<dimGrid, dimBlock, blockSize*sizeof(Float)>>>(A.data, A.pitch, A.nCols, square, result);
	QUIC_SAFE_CALL(cudaThreadSynchronize());
}

template <typename Float, QUIC::Uint blockSize>
inline void minMax(Float * blockMin, Float * blockMax, const QUIC::Uint& n)
{
	dim3 dimGrid(1);
	dim3 dimBlock(blockSize);
	minMaxKernel<Float, blockSize><<<dimGrid, dimBlock>>>(blockMin, blockMax, n);
	QUIC_SAFE_CALL(cudaThreadSynchronize());
}

template <typename Float>
void gpuCosines(const QUIC::MatrixGPU<Float>& A, Float * ARowNorms, QUIC::Uint * rows, const QUIC::Uint& nRows, const QUIC::Uint& splitRowId, Float * result, Float& cMin, Float& cMax, Float * minBuffer, Float * maxBuffer)
{
	const QUIC::Uint blockSizeX = 16;
	const QUIC::Uint blockSizeY = 16;
	dim3 dimBlock(blockSizeX, blockSizeY);
	dim3 dimGrid(1, static_cast<unsigned int>(iDivUp(nRows, blockSizeY)));
	const Float * splitRow = A.row(splitRowId);
	Float splitRowNorm;
	START_TIMER(COPY_H2D, GPU);
	QUIC_SAFE_CALL(cudaMemcpy(&splitRowNorm, ARowNorms + splitRowId, sizeof(Float), cudaMemcpyDeviceToHost));
	STOP_TIMER(COPY_H2D, GPU);
	cosineKernel<Float><<<dimGrid, dimBlock, blockSizeX*blockSizeY*sizeof(QUIC::Uint)>>>(A.data, A.pitch, ARowNorms, rows, nRows, A.nCols, splitRow, splitRowNorm, result, minBuffer, maxBuffer);
	QUIC_SAFE_CALL(cudaThreadSynchronize());
	QUIC::Uint nBlocks = dimGrid.x*dimGrid.y;
	if (nBlocks > 64)
	{
		if (nBlocks > 256) minMax<Float, 512>(minBuffer, maxBuffer, nBlocks); 
		else if (nBlocks > 128) minMax<Float, 256>(minBuffer, maxBuffer, nBlocks); 
		else minMax<Float, 128>(minBuffer, maxBuffer, nBlocks);
		QUIC_SAFE_CALL(cudaMemcpy(&cMin, minBuffer, sizeof(Float), cudaMemcpyDeviceToHost));
		QUIC_SAFE_CALL(cudaMemcpy(&cMax, maxBuffer, sizeof(Float), cudaMemcpyDeviceToHost));
	}
	else
	{
		Float tempMin [64];
		Float tempMax [64];
		QUIC_SAFE_CALL(cudaMemcpy(tempMin, minBuffer, nBlocks*sizeof(Float), cudaMemcpyDeviceToHost));
		QUIC_SAFE_CALL(cudaMemcpy(tempMax, maxBuffer, nBlocks*sizeof(Float), cudaMemcpyDeviceToHost));
		for (QUIC::Uint i=1; i<nBlocks; ++i)
		{
			if (tempMin[0] > tempMin[i]) tempMin[0] = tempMin[i];
			if (tempMax[0] < tempMax[i]) tempMax[0] = tempMax[i];
		}
		cMin = tempMin[0];
		cMax = tempMax[0];
	}
}

template <typename Float>
QUIC::Uint gpuLabelRows(Float * cosines, const QUIC::Uint& nRows, const Float& cMin, const Float& cMax, QUIC::Uint * labels, QUIC::Uint& blockSize, QUIC::Uint& nBlocks)
{
	const unsigned bs = 8*sizeof(QUIC::Uint);
	blockSize = static_cast<QUIC::Uint>(bs);
	nBlocks = iDivUp(nRows, blockSize);
	dim3 dimGrid(static_cast<unsigned int>(nBlocks));
	dim3 dimBlock(bs, 1);
	labelRowsKernel<Float, bs><<<dimGrid, dimBlock>>>(cosines, nRows, cMin, cMax, labels);
	QUIC_SAFE_CALL(cudaThreadSynchronize());
	return nBlocks;
}

template <typename Float>
inline void sumBlockNorms(Float * blockNorms, QUIC::Uint nBlockNorms, QUIC::Uint pitch, Float& leftInvNorm, Float& rightInvNorm)
{
	Float * leftBlockNorms = blockNorms;
	Float * rightBlockNorms = arrayRow(blockNorms, 1, pitch);
	const QUIC::Uint cpuThreshold = 32;
	if (nBlockNorms > cpuThreshold)
	{
		if (nBlockNorms > 96)
		{
			if (nBlockNorms > 384) sumBlockNormsKernel<Float, 512><<<2, 512>>>(blockNorms, nBlockNorms, pitch); 
			else if (nBlockNorms > 192) sumBlockNormsKernel<Float, 256><<<2, 256>>>(blockNorms, nBlockNorms, pitch); 
			else sumBlockNormsKernel<Float, 128><<<2, 128>>>(blockNorms, nBlockNorms, pitch);
		}
		else
		{
			if (nBlockNorms > 48) sumBlockNormsKernel<Float, 64><<<2, 64>>>(blockNorms, nBlockNorms, pitch);
			else sumBlockNormsKernel<Float, 32><<<2, 32>>>(blockNorms, nBlockNorms, pitch);
		}
		QUIC_SAFE_CALL(cudaMemcpy(&leftInvNorm, leftBlockNorms, sizeof(Float), cudaMemcpyDeviceToHost));
		QUIC_SAFE_CALL(cudaMemcpy(&rightInvNorm, rightBlockNorms, sizeof(Float), cudaMemcpyDeviceToHost));
	}
	else
	{
		Float tempLeft [cpuThreshold];
		Float tempRight [cpuThreshold];
		QUIC_SAFE_CALL(cudaMemcpy(tempLeft, leftBlockNorms, nBlockNorms*sizeof(Float), cudaMemcpyDeviceToHost));
		QUIC_SAFE_CALL(cudaMemcpy(tempRight, rightBlockNorms, nBlockNorms*sizeof(Float), cudaMemcpyDeviceToHost));
		leftInvNorm = Rsqrt(cpuVectorSum(nBlockNorms, tempLeft));
		rightInvNorm = Rsqrt(cpuVectorSum(nBlockNorms, tempRight));
	}
}

template <typename Float>
inline void normalizeRows(Float * left, Float * right, const QUIC::Uint& nCols, const Float& leftInvNorm, const Float& rightInvNorm)
{
	const QUIC::Uint blockSize = 512;
	dim3 dimGrid(static_cast<unsigned int>(iDivUp(nCols, blockSize)), 2);
	dim3 dimBlock(blockSize);
	normalizeKernel<Float><<<dimGrid, dimBlock>>>(left, right, nCols, leftInvNorm, rightInvNorm);
	QUIC_SAFE_CALL(cudaThreadSynchronize());
}

template <typename Float>
void gpuRowMeans(const QUIC::MatrixGPU<Float>& A, QUIC::Uint * rows, const QUIC::Uint& nRows, QUIC::Uint * labels, Float * leftMean, Float * rightMean, Float * scratch, const QUIC::Uint& pitch)
{
	const QUIC::Uint blockSize = 128;
	dim3 dimGrid(static_cast<unsigned int>(iDivUp(A.nCols, blockSize)));
	dim3 dimBlock(blockSize);
	assert(dimGrid.x*sizeof(Float) < pitch);
	rowMeansKernel<Float, blockSize><<<dimGrid, dimBlock>>>(A.data, A.pitch, A.nCols, rows, nRows, labels, leftMean, rightMean, scratch, pitch);
	QUIC_SAFE_CALL(cudaThreadSynchronize());
	Float leftInvNorm, rightInvNorm;
	sumBlockNorms(scratch, dimGrid.x, pitch, leftInvNorm, rightInvNorm);
	normalizeRows(leftMean, rightMean, A.nCols, leftInvNorm, rightInvNorm);
}

#include "explicit.h"
