#pragma once

#include <mkl_cblas.h>
#include <mkl_lapack.h>
#include <cublas.h>
#include <cutil_inline.h>
#include <cula.h>
#include <iostream>
#include <string>
#include <sstream>
#include <new>
#include "QUICException.h"
#include "QUIC.h"
#include "../debugPrint.h"

static QUIC::Uint nSampleMultiplier = 1;

#define QUIC_SAFE_NEW(className, ...) new (__FILE__, __LINE__, #className) className(__VA_ARGS__)

// For tracking device memory allocation/deallocation.
void devAllocTrack(void * p, QUIC::Uint n, char * file, int line);

void devFreeTrack(void * p, char * file, int line);

// sets the error flag and error message
inline void setError(const std::string& msg)
{
	extern std::string errorMsg;
	extern bool error;
	errorMsg = msg;
	error = true;
}

// error checking for kernels
inline bool QUICSafeCall(cudaError err, const char * file, const QUIC::Int line)
{
	if (err != cudaSuccess)
	{
		throw (QUICException(file, line, err));
	}
	return true;
}

#define QUIC_SAFE_CALL(err) QUICSafeCall(err, __FILE__, __LINE__)

template <typename Float>
QUIC::Int minArch();

template <>
inline QUIC::Int minArch<float>()
{
	return 10;
}

template <>
inline QUIC::Int minArch<double>()
{
	return 13;
}

#define INSTANTIATE_CLASS_TEMPLATE(CLASSNAME)	\
template CLASSNAME ## GPU<float>;				\
template CLASSNAME ## GPU<double>;				\
template CLASSNAME ## <float>;					\
template CLASSNAME ## <double>;		

// for svd
#define CPU_FULL "A"
#define GPU_FULL 'A'

// for matrix-matrix and matrix-vector products
#define CPU_TRANS_TYPE CBLAS_TRANSPOSE
#define GPU_TRANS_TYPE char
#define CPU_TRANS CblasTrans
#define GPU_TRANS 'T'
#define CPU_NO_TRANS CblasNoTrans
#define GPU_NO_TRANS 'N'

// square root
inline __host__ __device__ float Sqrt(const float& x) {return sqrtf(x);}
inline __host__ __device__ double Sqrt(const double& x) {return sqrt(x);}

// absolute value
inline __host__ __device__ float Abs(const float& x) {return fabsf(x);}
inline __host__ __device__ double Abs(const double& x) {return fabs(x);}

// vector-scalar products
inline void cpuScalarMul(const QUIC::Uint& n, const float& alpha, float * x) {cblas_sscal(static_cast<MKL_INT>(n), alpha, x, 1);}
inline void cpuScalarMul(const QUIC::Uint& n, const double& alpha, double * x) {cblas_dscal(static_cast<MKL_INT>(n), alpha, x, 1);}
inline void gpuScalarMul(const QUIC::Uint& n, const float& alpha, float * x) {cublasSscal(static_cast<int>(n), alpha, x, 1);}
inline void gpuScalarMul(const QUIC::Uint& n, const double& alpha, double * x) {cublasDscal(static_cast<int>(n), alpha, x, 1);}

// dot products
inline float cpuDot(const QUIC::Uint& n, const float * x, const float * y) {return cblas_sdot(static_cast<MKL_INT>(n), x, 1, y, 1);}
inline double cpuDot(const QUIC::Uint& n, const double * x, const double * y) {return cblas_ddot(static_cast<MKL_INT>(n), x, 1, y, 1);}
inline float gpuDot(const QUIC::Uint& n, const float * x, const float * y) {return cublasSdot(static_cast<int>(n), x, 1, y, 1); }
inline double gpuDot(const QUIC::Uint& n, const double * x, const double * y) {return cublasDdot(static_cast<int>(n), x, 1, y, 1); }

// vector sums
inline void cpuVectorSum(const QUIC::Uint& n, const float& alpha, const float * x, float * y) {cblas_saxpy(static_cast<MKL_INT>(n), alpha, x, 1, y, 1);}
inline void cpuVectorSum(const QUIC::Uint& n, const double& alpha, const double * x, double * y) {cblas_daxpy(static_cast<MKL_INT>(n), alpha, x, 1, y, 1);}
inline void gpuVectorSum(const QUIC::Uint& n, const float& alpha, const float * x, float * y) {cublasSaxpy(static_cast<int>(n), alpha, x, 1, y, 1);}
inline void gpuVectorSum(const QUIC::Uint& n, const double& alpha, const double * x, double * y) {cublasDaxpy(static_cast<int>(n), alpha, x, 1, y, 1);}

// matrix multiplication
inline void cpuMatMul(const CPU_TRANS_TYPE& transa, const CPU_TRANS_TYPE& transb, const QUIC::Uint& m, const QUIC::Uint& n, const QUIC::Uint& k, const float& alpha, const float * A, const QUIC::Uint& lda, const float * B, const QUIC::Uint& ldb, const float& beta, float * C, const QUIC::Uint& ldc) {cblas_sgemm(CblasColMajor, transa, transb, static_cast<MKL_INT>(m), static_cast<MKL_INT>(n), static_cast<MKL_INT>(k), alpha, A, static_cast<MKL_INT>(lda), B, static_cast<MKL_INT>(ldb), beta, C, static_cast<MKL_INT>(ldc));}
inline void cpuMatMul(const CPU_TRANS_TYPE& transa, const CPU_TRANS_TYPE& transb, const QUIC::Uint& m, const QUIC::Uint& n, const QUIC::Uint& k, const double& alpha, const double * A, const QUIC::Uint& lda, const double * B, const QUIC::Uint& ldb, const double& beta, double * C, const QUIC::Uint& ldc) {cblas_dgemm(CblasColMajor, transa, transb, static_cast<MKL_INT>(m), static_cast<MKL_INT>(n), static_cast<MKL_INT>(k), alpha, A, static_cast<MKL_INT>(lda), B, static_cast<MKL_INT>(ldb), beta, C, static_cast<MKL_INT>(ldc));}
inline void gpuMatMul(const GPU_TRANS_TYPE& transa, const GPU_TRANS_TYPE& transb, const QUIC::Uint& m, const QUIC::Uint& n, const QUIC::Uint& k, const float& alpha, const float * A, const QUIC::Uint& lda, const float * B, const QUIC::Uint& ldb, const float& beta, float * C, const QUIC::Uint& ldc) {cublasSgemm(transa, transb, static_cast<int>(m), static_cast<int>(n), static_cast<int>(k), alpha, A, static_cast<int>(lda), B, static_cast<int>(ldb), beta, C, static_cast<int>(ldc));}
inline void gpuMatMul(const GPU_TRANS_TYPE& transa, const GPU_TRANS_TYPE& transb, const QUIC::Uint& m, const QUIC::Uint& n, const QUIC::Uint& k, const double& alpha, const double * A, const QUIC::Uint& lda, const double * B, const QUIC::Uint& ldb, const double& beta, double * C, const QUIC::Uint& ldc) {cublasDgemm(transa, transb, static_cast<int>(m), static_cast<int>(n), static_cast<int>(k), alpha, A, static_cast<int>(lda), B, static_cast<int>(ldb), beta, C, static_cast<int>(ldc));}

// svd
inline bool gpuSVD(const char& jobu, const char& jobvt, const QUIC::Uint& m, const QUIC::Uint& n, float * A, const QUIC::Uint& lda, float * S, float * U, const QUIC::Uint& ldu, float * VT, const QUIC::Uint& ldvt)
{
	culaStatus s = culaDeviceSgesvd(jobu, jobvt, static_cast<int>(m), static_cast<int>(n), A, static_cast<int>(lda), S, U, static_cast<int>(ldu), VT, static_cast<int>(ldvt));
	if (s != culaNoError)
	{
		throw (QUICException(__FILE__, __LINE__, s));
	}
	return true;
}

inline bool gpuSVD(const char& jobu, const char& jobvt, const QUIC::Uint& m, const QUIC::Uint& n, double * A, const QUIC::Uint& lda, double * S, double * U, const QUIC::Uint& ldu, double * VT, const QUIC::Uint& ldvt)
{
#if defined(CULA_DOUBLE_PRECISION) //Define this if you have double precision cula and can run culaDeviceDgesvd
	culaStatus s = culaDeviceDgesvd(jobu, jobvt, static_cast<int>(m), static_cast<int>(n), A, static_cast<int>(lda), S, U, static_cast<int>(ldu), VT, static_cast<int>(ldvt));
#else
	culaStatus s = culaDataError; //This is a bit rough
#endif
	if (s != culaNoError)
	{
		throw (QUICException(__FILE__, __LINE__, s));
	}
	return false;
}

inline bool cpuSVD(char * jobu, char * jobvt, QUIC::Uint m, QUIC::Uint n, float * A, QUIC::Uint lda, float * S, float * U, QUIC::Uint ldu, float * VT, QUIC::Uint ldvt, float * work, QUIC::Uint lwork) 
{
	MKL_INT info = 0LL;
	MKL_INT mklM = static_cast<MKL_INT>(m);
	MKL_INT mklN = static_cast<MKL_INT>(n);
	MKL_INT mklLda = static_cast<MKL_INT>(lda);
	MKL_INT mklLdu = static_cast<MKL_INT>(ldu);
	MKL_INT mklLdvt = static_cast<MKL_INT>(ldvt);
	MKL_INT mklLwork = static_cast<MKL_INT>(lwork);
	sgesvd(jobu, jobvt, &mklM, &mklN, A, &mklLda, S, U, &mklLdu, VT, &mklLdvt, work, &mklLwork, &info);
	if (info != 0LL)
	{
		std::stringstream ss;
		if (info < 0) ss<<"sgesvd: illegal value for parameter "<<-info;
		else ss<<"sgesvd: failed to converge.";
		throw(QUICException(__FILE__, __LINE__, ss.str().c_str()));
	}
	return true;
}

inline bool cpuSVD(char * jobu, char * jobvt, QUIC::Int m, QUIC::Int n, double * A, QUIC::Int lda, double * S, double * U, QUIC::Int ldu, double * VT, QUIC::Int ldvt, double * work, QUIC::Int lwork) 
{
	MKL_INT info = 0LL;
	MKL_INT mklM = static_cast<MKL_INT>(m);
	MKL_INT mklN = static_cast<MKL_INT>(n);
	MKL_INT mklLda = static_cast<MKL_INT>(lda);
	MKL_INT mklLdu = static_cast<MKL_INT>(ldu);
	MKL_INT mklLdvt = static_cast<MKL_INT>(ldvt);
	MKL_INT mklLwork = static_cast<MKL_INT>(lwork);
	dgesvd(jobu, jobvt, &mklM, &mklN, A, &mklLda, S, U, &mklLdu, VT, &mklLdvt, work, &mklLwork, &info);
	if (info != 0LL)
	{
		std::stringstream ss;
		if (info < 0) ss<<"dgesvd: illegal value for parameter "<<-info;
		else ss<<"dgesvd: failed to converge.";
		throw (QUICException(__FILE__, __LINE__, ss.str().c_str()));
	}
	return true;
}

// vector norm
inline float cpuNorm(const QUIC::Uint& n, const float * x) {return cblas_snrm2(static_cast<MKL_INT>(n), x, 1);}
inline double cpuNorm(const QUIC::Uint& n, const double * x) {return cblas_dnrm2(static_cast<MKL_INT>(n), x, 1);}
inline float gpuNorm(const QUIC::Uint& n, const float * x) {return cublasSnrm2(static_cast<int>(n), x, 1);}
inline double gpuNorm(const QUIC::Uint& n, const double * x) {return cublasDnrm2(static_cast<int>(n), x, 1);}

// vector sum
inline float cpuVectorSum(const QUIC::Uint& n, float * x) {return cblas_sasum(static_cast<MKL_INT>(n), x, 1);}
inline double cpuVectorSum(const QUIC::Uint& n, double * x) {return cblas_dasum(static_cast<MKL_INT>(n), x, 1);}

// replaces each element of m with it's square root
template <typename Float>
void matrixSqrt(QUIC::MatrixGPU<Float>& m);

template <typename Float>
void matrixSqrt(QUIC::Matrix<Float>& m);

// computes A/diag(v) and stores the result in B.
template <typename Float>
void diagDivR(const QUIC::MatrixGPU<Float>& A, const QUIC::MatrixGPU<Float>& v, QUIC::MatrixGPU<Float>& B);

template <typename Float>
void diagDivR(const QUIC::Matrix<Float>& A, const QUIC::Matrix<Float>& v, QUIC::Matrix<Float>& B);

// computes diag(v)\A and stores the result in B.
template <typename Float>
void diagDivL(const QUIC::MatrixGPU<Float>& A, const QUIC::MatrixGPU<Float>& v, QUIC::MatrixGPU<Float>& B);

template <typename Float>
void diagDivL(const QUIC::Matrix<Float>& A, const QUIC::Matrix<Float>& v, QUIC::Matrix<Float>& B);

// elementwise vector products
template <typename Float>
void vectorMul(const QUIC::MatrixGPU<Float>& x, QUIC::MatrixGPU<Float>& y);

template <typename Float>
void vectorMul(const QUIC::Matrix<Float>& x, QUIC::Matrix<Float>& y);

// computes the squared norms of all rows in a matrix
template <typename Float>
void rowNorms(const QUIC::MatrixGPU<Float>& A, const bool& square, Float * result);

template <typename Float>
void rowNorms(const QUIC::Matrix<Float>& A, const bool& square, Float * result);

// computes the cosines of the angles betweenthe rows of A index by the array rows and the split row
template <typename Float>
void gpuCosines(const QUIC::MatrixGPU<Float>& A, Float * ARowNorms, QUIC::Uint * rows, const QUIC::Uint& nRows, const QUIC::Uint& splitRowId, Float * result, Float& cMin, Float& cMax, Float * minBuffer, Float * maxBuffer);

// computes the label (left or right) for every row
template <typename Float>
QUIC::Uint gpuLabelRows(Float * cosines, const QUIC::Uint& nRows, const Float& cMin, const Float& cMax, QUIC::Uint * labels, QUIC::Uint& blockSize, QUIC::Uint& nBlocks);

template <typename Float>
QUIC::Uint cpuLabelRows(Float * cosines, const QUIC::Uint& nRows, const Float& cMin, const Float& cMax, QUIC::Uint * labels, QUIC::Uint& blockSize, QUIC::Uint& nBlocks);

// finds the means of the rows indexed by the array rows.
template <typename Float>
void gpuRowMeans(const QUIC::MatrixGPU<Float>& A, QUIC::Uint * rows, const QUIC::Uint& nRows, QUIC::Uint * labels, Float * leftMean, Float * rightMean, Float * scratch, const QUIC::Uint& scratchPitch); 

// integer division, rounding up
inline QUIC::Uint iDivUp(const QUIC::Uint& a, const QUIC::Uint& b)
{
	return a/b + (a%b ? 1 : 0);
}

template <typename Matrix, typename Float>
bool extractSVDFromProjection(const Matrix& AVhat, const Matrix& VhatT, Matrix * U, Matrix * S, Matrix * VT, Float * scratch, const size_t& scratchPitch);

// inverse of thecontinuous distribution function of the normal distrubution.
double ltqnorm(double p);

inline double icdf(const double& x, const double& mean, const double& variance)
{
	DEBUG_PRINT("%f\n", ltqnorm(x)*Sqrt(variance));
	return ltqnorm(x)*Sqrt(variance) + mean;
}

// Memory allocation functions
#define QUIC_SAFE_ALLOC(T, n) hostAlloc< T >(n, __FILE__, __LINE__, #T)
#define QUIC_SAFE_ALLOC_2D(T, w, h, p) hostAlloc< T >(w, h, p, __FILE__, __LINE__, #T)

template <typename T>
inline T * hostAlloc(const QUIC::Uint& n, const char *file, const QUIC::Int line, char * typeName)
{
	if (n == 0) return NULL;
	T * mem;
	try
	{
		mem = reinterpret_cast<T *>(::operator new(n*sizeof(T)));
	}
	catch(std::bad_alloc&)
	{
		std::stringstream ss;
		ss<<"Memory allocation for type "<<typeName<<" ["<<(n/sizeof(T))<<"] failed."<<std::endl;
		throw(QUICException(file, line, ss.str().c_str()));
	}
	return mem;
}

template <typename T>
inline T * hostAllocMapped(const QUIC::Uint& n)
{
	T * t;
	if (QUIC_SAFE_CALL(cudaHostAlloc(reinterpret_cast<void **>(&t), n*sizeof(T), cudaHostAllocMapped))) return t;
	return NULL;
}

inline void hostFree(void * p)
{
	::operator delete(p);
}

template <typename T>
inline T * devAlloc(const QUIC::Uint& n)
{
	T * t;
	if (!QUIC_SAFE_CALL(cudaMalloc(reinterpret_cast<void **>(&t), n*sizeof(T)))) t = NULL;
	devAllocTrack(t, n, __FILE__, __LINE__);
	return t;
}

template <typename T>
inline T * devAlloc2D(const QUIC::Uint& w, const QUIC::Uint& h, QUIC::Uint& pitch)
{
	extern size_t maxPitch;
	T * t;
	QUIC::Uint rowWidth = w*sizeof(T);
	if (iDivUp(rowWidth, 256)*256 > maxPitch)
	{
		devAllocTrack(NULL, h*iDivUp(rowWidth, 256)*256, __FILE__, __LINE__);
		throw (QUICException(__FILE__, __LINE__, "Maximum pitch exceeded."));
	}
	size_t p;
	bool result = QUIC_SAFE_CALL(cudaMallocPitch(reinterpret_cast<void **>(&t), &p, rowWidth, h));
	pitch = static_cast<QUIC::Uint>(p);
	devAllocTrack(t, h*p, __FILE__, __LINE__);
	return result ? t : NULL;
}

template <typename T>
inline void devFree(T * t)
{
	devFreeTrack(t, __FILE__, __LINE__);
	QUIC_SAFE_CALL(cudaFree(t));
}

// host to host and device to device memory copy
template <typename T>
inline void cpuMemcpy(const QUIC::Uint& n, const T * x, T * y)
{
	memcpy(y, x, n*sizeof(T));
}

template <typename T>
inline bool gpuMemcpy(const QUIC::Uint& n, const T * x, T * y)
{
	return QUIC_SAFE_CALL(cudaMemcpy(y, x, n*sizeof(T), cudaMemcpyDeviceToDevice));
}

// retrurns a pointer to the desired row of a 2D pitched array
template <typename T>
inline __host__ __device__ T * arrayRow(T * A, const QUIC::Uint& r, const QUIC::Uint& pitch)
{
	return reinterpret_cast<T *>(reinterpret_cast<char *>(A) + r*pitch);
}

template <typename T>
inline __host__ __device__ const T * arrayRow(const T * A, const QUIC::Uint& r, const QUIC::Uint& pitch)
{
	return reinterpret_cast<const T *>(reinterpret_cast<const char *>(A) + r*pitch);
}

template <typename T>
inline __host__ __device__ T square(const T& t)
{
	return t*t;
}

template <typename Float> 
inline bool checkArch(const QUIC::Int& gpuArch)
{
	if (gpuArch < minArch<Float>())
	{
		std::stringstream ss; 
		ss<<"GPU architecture "<<minArch<Float>()<<" or higher is required."<<std::endl;
		setError(ss.str());
		return false;
	}
	return true;
}

#if defined __INTEL_COMPILER

// returns the ceiling of the log base 2 of an integer, i.e. the mimimum number of bits needed to store it
inline QUIC::Uint log2(QUIC::Uint x)
{
	QUIC::Uint result;
	__asm
	{
		BSR RAX, x
		MOV result, RAX
	}
	return result;
}

// counts the total number of bits in an array of integers.
inline QUIC::Uint countBits(QUIC::Uint * x, QUIC::Uint n)
{
	QUIC::Uint count = 0;
	__asm
	{
		  MOV RBX, x
		  MOV RDX, n
		  MOV RCX, 0
	LOOP: MOV RAX, [RBX]
		  POPCNT RAX, RAX
		  ADD RCX, RAX
		  ADD RBX, 8
		  DEC RDX
		  JNZ LOOP
		  MOV count, RCX
	}
	return count;
}

#elif defined _MSC_VER

#include <intrin.h>
#include <nmmintrin.h>

// returns the ceiling of the log base 2 of an integer, i.e. the mimimum number of bits needed to store it
inline QUIC::Uint log2(QUIC::Uint x)
{
	unsigned long result;
	_BitScanReverse64(&result, x);
	return static_cast<QUIC::Uint>(result);
}

// counts the total number of bits in an array of integers.
inline QUIC::Uint countBits(QUIC::Uint * x, QUIC::Uint n)
{
	QUIC::Uint count = 0;
	for (QUIC::Uint i=0; i<n; ++i) count += _mm_popcnt_u64(x[i]);
	return count;
}

#endif

inline QUIC::Uint nSamples(const QUIC::Uint& nRows)
{
	QUIC::Uint s = nSampleMultiplier*log2(nRows);
	if (s > nRows) s = nRows;
	if (s < 1) s = 1;
	return s;
}