template <typename Float>
__global__ void vectorSqrtKernel(Float * m, QUIC::Uint pitch, QUIC::Uint nRows, QUIC::Uint nCols)
{
	QUIC::Uint idx = blockIdx.x*blockDim.x + threadIdx.x;
	QUIC::Uint idy = blockIdx.y*blockDim.y + threadIdx.y;
	if (idx < nCols && idy < nRows)
	{
		Float * row = arrayRow(m, idy, pitch);
		Float * entry = row + idx;
		*entry = Sqrt(*entry);
	}
}

template <typename Float, QUIC::Uint blockSizeX>
__global__ void diagDivRKernel(Float * A, QUIC::Uint Apitch, Float * v, Float * B, QUIC::Uint Bpitch, QUIC::Uint nRows, QUIC::Uint nCols)
{
	__shared__ Float vComp [blockSizeX];
	const QUIC::Uint blockX = blockIdx.x*blockDim.x;
	const QUIC::Uint blockY = blockIdx.y*blockDim.y;
	const QUIC::Uint x = blockX + threadIdx.x;
	const QUIC::Uint y = blockY + threadIdx.y;
	const QUIC::Uint threadId = threadIdx.y*blockDim.x + threadIdx.x;
	if (threadId < blockSizeX && blockX + threadId < nCols)
	{
		vComp[threadId] = v[blockX + threadId];
	}
	__syncthreads();
	if (x < nCols && y < nRows)
	{
		Float vElt = vComp[threadIdx.x];
		const Float * Aelt = arrayRow(A, y, Apitch) + x;
		Float * Belt = arrayRow(B, y, Bpitch) + x;
		*Belt = *Aelt/vElt;
	}
}

template <typename Float, QUIC::Uint blockSizeY>
__global__ void diagDivLKernel(const Float * A, QUIC::Uint Apitch, const Float * v, Float * B, QUIC::Uint Bpitch, QUIC::Uint nRows, QUIC::Uint nCols)
{
	__shared__ Float vComp [blockSizeY];
	const QUIC::Uint blockX = blockIdx.x*blockDim.x;
	const QUIC::Uint blockY = blockIdx.y*blockDim.y;
	const QUIC::Uint x = blockX + threadIdx.x;
	const QUIC::Uint y = blockY + threadIdx.y;
	const QUIC::Uint threadId = threadIdx.y*blockDim.x + threadIdx.x;
	if (threadId < blockSizeY && blockY + threadId < nRows)
	{
		vComp[threadId] = v[blockY + threadId];
	}
	__syncthreads();
	if (x < nCols && y < nRows)
	{
		Float vElt = vComp[threadIdx.y];
		const Float * Aelt = arrayRow(A, y, Apitch) + x;
		Float * Belt = arrayRow(B, y, Bpitch) + x;
		*Belt = *Aelt/vElt;
	}
}

template <typename Float>
__global__ void vectorMulKernel(Float * x, Float * y, QUIC::Uint n)
{
	QUIC::Uint id = blockIdx.x*blockDim.x + threadIdx.x;
	if (id < n) y[id] *= x[id];
}

template <typename Float>
__device__ Float * smem();

template <>
__device__ float * smem<float>()
{
	extern __shared__ float smemf [];
	return smemf;
}

template <>
__device__ double * smem<double>()
{
	extern __shared__ double smemd [];
	return smemd;
}

template <typename Float>
__global__ void sqRowNormsKernel(const Float * A, QUIC::Uint pitch, QUIC::Uint nCols, const bool square, Float * result)
{
	Float * sharedAcc = smem<Float>();
	const Float * row = arrayRow(A, blockIdx.x, pitch);
	Float regAcc = 0.0f;
	for (QUIC::Uint i=threadIdx.x; i<nCols; i+=blockDim.x)
	{
		Float entry = row[i];
		regAcc += entry*entry;
	}
	sharedAcc[threadIdx.x] = regAcc;
	for(QUIC::Uint stride = blockDim.x >> 1; stride > 1; stride >>= 1)
	{
		__syncthreads();
		if (threadIdx.x < stride)
		{
			Float x = sharedAcc[threadIdx.x + stride];
			sharedAcc[threadIdx.x] += x;
		}
	}
	__syncthreads();
	if (threadIdx.x == 0)
	{
		Float sum = sharedAcc[0] + sharedAcc[1];
		result[blockIdx.x] = square ? sum : Sqrt(sum);
	}
}

template <typename Float>
__global__ void cosineKernel(const Float * A, QUIC::Uint pitch, Float * ARowNorms, QUIC::Uint * rows, const QUIC::Uint nRows, const QUIC::Uint nCols, const Float * splitRow, Float splitRowNorm, Float * result, Float * blockMin, Float * blockMax)
{
	const QUIC::Uint y0 = blockIdx.y*blockDim.y;
	const QUIC::Uint threadId = blockDim.x*threadIdx.y + threadIdx.x;
	const QUIC::Uint y1 = y0 + threadIdx.y;
	const QUIC::Uint y2 = y0 + threadId;
	const QUIC::Uint blockId = blockIdx.y*gridDim.x + blockIdx.x;
	extern __shared__ char smem [];
	Float * const sharedAcc = reinterpret_cast<Float *>(smem) + threadIdx.y*blockDim.x;
	Float * const sharedAcc2 = reinterpret_cast<Float *>(smem) + threadId*blockDim.x;
	Float * const& sharedSplitRow = reinterpret_cast<Float *>(smem);
	Float * const& sharedMinCosines = reinterpret_cast<Float *>(smem);
	Float * const sharedMaxCosines = sharedMinCosines + blockDim.y;
	QUIC::Uint * const& sharedRowIds = reinterpret_cast<QUIC::Uint *>(smem);
	// read the row id's and row norms into shared memory
	Float rowNorm;
	if (y2 < nRows && threadId < blockDim.y)
	{		
		QUIC::Uint rowId = rows[y2];
		sharedRowIds[threadId] = rowId;
		rowNorm = ARowNorms[rowId];
	}
	__syncthreads();
	// read the row id's and row norms from shared memorey
	const Float * row = arrayRow(A, sharedRowIds[threadIdx.y], pitch);
	__syncthreads();
	// each thread accumulates a subset of the terms of a dot product
	Float regAcc = 0.0f;
	for (QUIC::Uint i=0; i<nCols; i+=blockDim.x)
	{
		QUIC::Uint c = i + threadIdx.x;
		if (threadIdx.y == 0) sharedSplitRow[threadIdx.x] = c < nCols ? splitRow[c] : 0.0f;
		__syncthreads();
		if (y1 < nRows) regAcc += row[c]*sharedSplitRow[threadIdx.x];
		__syncthreads();
	}
	// use parallel reduction to sum the per-thread accumulators
	sharedAcc[threadIdx.x] = regAcc;
	for(QUIC::Uint stride = blockDim.x >> 1; stride > 1; stride >>= 1)
	{
		__syncthreads();
		if (threadIdx.x < stride) sharedAcc[threadIdx.x] += sharedAcc[threadIdx.x + stride];
	}
	__syncthreads();
	// write the result back to global memory and prepare for min/max computation
	Float cMin = 1.0f;
	Float cMax = 0.0f;
	if (threadId < blockDim.y && y2 < nRows)
	{
		Float cosine = Abs((sharedAcc2[0] + sharedAcc2[1])/(rowNorm*splitRowNorm));
		result[y2] = cosine;
		cMin = cosine;
		cMax = cosine;
	}
	// initialize shared min/max arrays
	__syncthreads();
	if (threadId < blockDim.y)
	{
		sharedMinCosines[threadId] = cMin;
		sharedMaxCosines[threadId] = cMax;
	}
	__syncthreads();
	// find min/max via parallel reduction
	for(QUIC::Uint stride = blockDim.y >> 1; stride > 0; stride >>= 1)
	{
		__syncthreads();
		if (threadId < stride)
		{
			Float cMin2 = sharedMinCosines[threadId + stride];
			Float cMax2 = sharedMaxCosines[threadId + stride];
			if (cMin2 < cMin)
			{
				cMin = cMin2;
				sharedMinCosines[threadId] = cMin;
			}
			if (cMax2 > cMax)
			{
				cMax = cMax2;
				sharedMaxCosines[threadId] = cMax;
			}
		}
	}
	__syncthreads();
	// write the block min and max to global memory
	if (threadId == 0)
	{
		blockMin[blockId] = cMin;
		blockMax[blockId] = cMax;
	}
}

template <typename Float, QUIC::Uint blockSize>
__global__ void minMaxKernel(Float * blockMin, Float * blockMax, QUIC::Uint n)
{
	__shared__ Float sharedMin [blockSize];
	__shared__ Float sharedMax [blockSize];
	Float cMin = 1.0f;
	Float cMax = 0.0f;
	for (QUIC::Uint i=threadIdx.x; i<n; i+=blockSize)
	{
		Float cMin2 = blockMin[i];
		Float cMax2 = blockMax[i];
		if (cMin2 < cMin) cMin = cMin2;
		if (cMax2 > cMax) cMax = cMax2;
	}
	sharedMin[threadIdx.x] = cMin;
	sharedMax[threadIdx.x] = cMax;
	for(QUIC::Uint stride = blockDim.x >> 1; stride > 0; stride >>= 1)
	{
		__syncthreads();
		if (threadIdx.x < stride)
		{
			Float cMin2 = sharedMin[threadIdx.x + stride];
			if (cMin2 < cMin)
			{
				cMin = cMin2;
				sharedMin[threadIdx.x] = cMin;
			}
			Float cMax2 = sharedMax[threadIdx.x + stride];
			if (cMax2 > cMax)
			{
				cMax = cMax2;
				sharedMax[threadIdx.x] = cMax;
			}
		}
	}
	__syncthreads();
	if (threadIdx.x == 0)
	{
		*blockMin = cMin;
		*blockMax = cMax;
	}
}

template <typename Float, QUIC::Uint blockSize>
__global__ void labelRowsKernel(Float * cosines, QUIC::Uint nRows, Float cMin, Float cMax, QUIC::Uint * labels)
{
	__shared__ QUIC::Uint blockLabels [blockSize];
	const QUIC::Uint id = blockIdx.x*blockDim.x + threadIdx.x;
	QUIC::Uint label = 0;
	if (id < nRows)
	{
		Float cosine = cosines[id];
		if (cMax - cosine <= cosine - cMin) label = static_cast<QUIC::Uint>(1) << threadIdx.x;
	}
	blockLabels[threadIdx.x] = label;
	for(QUIC::Uint stride = blockDim.x >> 1; stride > 0; stride >>= 1)
	{
		__syncthreads();
		if (threadIdx.x < stride) blockLabels[threadIdx.x] |= blockLabels[threadIdx.x + stride];
	}
	__syncthreads();
	if (threadIdx.x == 0) labels[blockIdx.x] = blockLabels[0];
}

template <typename Float, QUIC::Uint blockSize>
__global__ void sumBlockNormsKernel(Float * blockNorms, QUIC::Uint nBlockNorms, QUIC::Uint pitch)
{
	__shared__ Float sharedAcc [blockSize];
	Float acc = 0.0f;
	Float * row = arrayRow(blockNorms, blockIdx.x, pitch);
	for (QUIC::Uint i=threadIdx.x; i<nBlockNorms; i+=blockDim.x)
	{
		acc += row[i];
	}
	sharedAcc[threadIdx.x] = acc;
	__syncthreads();
	for (QUIC::Uint stride = blockDim.x >> 1; stride > 0; stride >>= 1)
	{
		__syncthreads();
		if (threadIdx.x < stride) sharedAcc[threadIdx.x] += sharedAcc[threadIdx.x + stride];
	}
	__syncthreads();
	if (threadIdx.x == 0) row[0] = Rsqrt(sharedAcc[0]);
}

template <typename Float>
__global__ void normalizeKernel(Float * left, Float * right, QUIC::Uint nCols, Float leftInvNorm, Float rightInvNorm)
{
	QUIC::Uint x = blockIdx.x*blockDim.x + threadIdx.x;
	if (x < nCols)
	{
		if (blockIdx.y == 0) left[x] *= leftInvNorm;
		else right[x] *= rightInvNorm;
	}
}

template <typename Float, QUIC::Uint blockSize>
__global__ void rowMeansKernel(Float * A, QUIC::Uint pitchA, QUIC::Uint nCols, QUIC::Uint * rows, QUIC::Uint nRows, QUIC::Uint * labels, Float * leftMean, Float * rightMean, Float * scratch, QUIC::Uint pitchS)
{
	__shared__ QUIC::Uint sharedRowId; 
	__shared__ QUIC::Uint sharedLabels;
	__shared__ Float sharedSums [blockSize];
	QUIC::Uint x = blockIdx.x*blockDim.x + threadIdx.x;
	Float acc []  = {0.0f, 0.0f};
	QUIC::Int lastLabelArrayIndex = -1;
	for (QUIC::Uint y=0; y<nRows; ++y)
	{
		QUIC::Int labelArrayIndex = y/static_cast<QUIC::Int>(8*sizeof(QUIC::Uint));
		QUIC::Int labelBitIndex = y - labelArrayIndex*8*sizeof(QUIC::Uint);
		// read the row id and bit mask into shared memory
		if (threadIdx.x == 0)
		{
			sharedRowId = rows[y];
			if (labelArrayIndex != lastLabelArrayIndex)
			{
				lastLabelArrayIndex = labelArrayIndex;
				sharedLabels = labels[labelArrayIndex];
			}
		}
		__syncthreads();
		// add to the accumulator
		if (x < nCols)
		{
			Float * row = arrayRow(A, sharedRowId, pitchA);
			QUIC::Uint label = (sharedLabels >> labelBitIndex) & 1;
			acc[label] += row[x];
		}
		__syncthreads();
	}
	// write the result to global memory
	if (x < nCols)
	{
		leftMean[x] =  acc[1];
		rightMean[x] = acc[0];
	}
	// compute the sum of squares of the block result with parallel reduction
	for (QUIC::Int i=0; i<2; ++i)
	{
		sharedSums[threadIdx.x] = acc[i]*acc[i];
		for(QUIC::Uint stride = blockDim.x >> 1; stride > 0; stride >>= 1)
		{
			__syncthreads();
			if (threadIdx.x < stride) sharedSums[threadIdx.x] += sharedSums[threadIdx.x + stride];
		}
		__syncthreads();
		if (threadIdx.x == 0) acc[i] = sharedSums[0];
		__syncthreads();
	}
	// write the block sum of squares to global memory
	if (threadIdx.x == 0)
	{
		Float * leftBlockNorms = scratch;
		Float * rightBlockNorms = arrayRow(scratch, 1, pitchS);
		leftBlockNorms[blockIdx.x] = acc[1];
		rightBlockNorms[blockIdx.x] = acc[0];
	}
}