#include "mex.h"
#include "QUIC.h"
#include <stdarg.h>
#include "Timer.h"
#include "MxArrayAllocator.h"
#include "errorFuncs.h"
#include "../debugPrint.h"

using namespace QUIC;

typedef enum {CPU, GPU} Version;

typedef enum {PROFILE, BLOCK_SIZE} Option;

template <typename T>
bool parseArg(int n, const char * argName, T& argValue, ...)
{
	va_list args;
	va_start(args, argValue);
	for (int i=0; i<n; ++i)
	{
		char * name = va_arg(args, char *);
		T value = va_arg(args, T);
		if (_stricmp(argName, name) == 0)
		{
			argValue = value;
			return true;
		}
	}
	return false;
}

template <typename Float>
void runQUIC(Version version, const mxArray * mxAT, mxArray ** mxUT, mxArray ** mxS, mxArray ** mxV, Float epsilon, Float delta, OrthoMode orthoMode, int maxRank, int partitions)
{
	if (mxS == NULL) return;
	MxArrayAllocator<Float> UTalloc, Salloc, Valloc;
	Matrix<Float> UT(&UTalloc), S(&Salloc), V(&Valloc);
	mexQUICError();
	size_t m = mxGetM(mxAT);
	size_t n = mxGetN(mxAT);
	if (m == 0 || n == 0)
	{
		if (mxUT) UTalloc.allocate(0, 0);
		Salloc.allocate(0, 0);
		if (mxV) Valloc.allocate(0, 0);	
	}
	else
	{
		Matrix<Float> AT(n, m, reinterpret_cast<Float *>(mxGetData(mxAT)), m*sizeof(Float));
		if (version == GPU)
		{
			//initGPU();
			//MatrixGPU<Float> ATgpu(AT);
			//MatrixGPU<Float> UTgpu, Sgpu, Vgpu;
			//mexQUICError();
			//QUIC_SVD<Float>(&ATgpu, mxV ? &Vgpu : NULL, &Sgpu, mxUT ? &UTgpu : NULL, epsilon, delta, orthoMode);
			//mexQUICError();
			//UT = UTgpu;
			//S = Sgpu;
			//V = Vgpu;
			//mexQUICError();
			//shutdownGPU();

			DEBUG_PRINT("Initializing GPU\n");
			initGPU();
			DEBUG_PRINT("Creating matrices\n");
			MatrixGPU<Float> UTgpu, Sgpu, Vgpu;
			mexQUICError();
			QUIC_SVD<Float>(&AT, mxV ? &Vgpu : NULL, &Sgpu, mxUT ? &UTgpu : NULL, epsilon, delta, orthoMode, maxRank, partitions);
			mexQUICError();
			DEBUG_PRINT("Copying result\n");
			UT = UTgpu;
			S = Sgpu;
			V = Vgpu;
			mexQUICError();
			//shutdownGPU();
		}
		else
		{
			//QUIC_SVD<Float>(&AT, mxV ? &V : NULL, &S, mxUT ? &UT : NULL, epsilon, delta, orthoMode);
			//mexQUICError();
			QUIC_SVD<Float>(&AT, mxV ? &V : NULL, &S, mxUT ? &UT : NULL, epsilon, delta, orthoMode, maxRank, partitions);
			mexQUICError();
		}
	}
	if (mxUT) *mxUT = UTalloc.A;
	*mxS = Salloc.A;
	if (mxV) *mxV = Valloc.A;
}

void mexFunction(int nlhs, mxArray * plhs[], int nrhs, const mxArray * prhs[])
{


		const int minArgs = 3;
		if (nrhs < minArgs) usageError();
		if ((nrhs - minArgs)%2) usageError();
		if (!mxIsChar(prhs[0])) usageError();
		const mxArray * mxVersion = prhs[0];
		const mxArray * mxOrthoMode = prhs[1];
		const mxArray * mxAT = prhs[2];
		mxArray ** mxUT = nlhs >= 2 ? plhs : NULL;
		mxArray ** mxS = nlhs == 0 ? NULL : nlhs == 1 ? plhs : plhs + 1;
		mxArray ** mxV = nlhs >= 3 ? plhs + 2 : NULL;

		const int bufferSize = 16;
		char buffer [bufferSize];
		Version version;
		OrthoMode orthoMode;
		int maxRank = 0;
		mxGetString(mxVersion, buffer, bufferSize);
		if (!parseArg<Version>(2, buffer, version, "CPU", CPU, "GPU", GPU)) usageError();
		mxGetString(mxOrthoMode, buffer, bufferSize);
		if (!parseArg<OrthoMode>(3, buffer, orthoMode, "MGS", MGS, "UGS", UGS, "BLOCK_MGS", BLOCK_MGS)) usageError();
		if (mxIsSparse(mxAT) || mxIsComplex(mxAT)) mexErrMsgTxt("Only dense real matrices are supported.");
		bool profile = false;
		double epsilon = 0.0001;
		double delta = 0.01;
		int partitions = 1;
		for (int i=minArgs; i<nrhs; i+=2)
		{
			mxGetString(prhs[i], buffer, bufferSize);
			string argName(buffer);
			for (string::iterator iter = argName.begin(); iter != argName.end(); ++iter) *iter = tolower(*iter);
			if (mxGetM(prhs[i + 1]) != 1 || mxGetN(prhs[i + 1]) != 1) invalidValue(argName);
			double argValue = mxGetScalar(prhs[i + 1]);
			QUIC::Uint uiArgValue = static_cast<QUIC::Uint>(argValue);
			if (argName == "profile")
			{
				if (uiArgValue != 1uLL && argValue != 0uLL) invalidValue(argName);
				profile = (argValue == 1uLL);
			}
			else if (argName == "blocksize")
			{
				if (orthoMode != BLOCK_MGS) mexErrMsgTxt("Invalid parameter 'blockSize'");
				if (argValue != static_cast<double>(uiArgValue) || uiArgValue < 1uLL) invalidValue(argName);
			}
			else if (argName == "epsilon")
			{
				if (argValue < 0.0) invalidValue(argName);
				epsilon = argValue;
			}
			else if (argName == "delta")
			{
				if (argValue < 0.0 || argValue > 1.0) invalidValue(argName);
				delta = argValue;
			}
			else if (argName == "partitions")
			{
				int val = static_cast<int>(argValue);
				if (val < 1) invalidValue(argName);
				partitions = val;
			}
			else if (argName == "maxrank")
			{
				int val = static_cast<int>(argValue);
				if (val < 0) invalidValue(argName);
				maxRank = val;
			}
			else
			{
				stringstream ss;
				ss<<"Invalid parameter '"<<argName<<"'";
				mexErrMsgTxt(ss.str().c_str());
			}
		}
		if (profile) resetTimers();



		if (mxIsSingle(mxAT)) 
		{
			runQUIC<float>(version, mxAT, mxUT, mxS, mxV, static_cast<float>(epsilon), static_cast<float>(delta), orthoMode, partitions, maxRank);
		}
		else if (mxIsDouble(mxAT)) 
		{
			runQUIC<double>(version, mxAT, mxUT, mxS, mxV, epsilon, delta, orthoMode, partitions, maxRank);
		}
		else 
			mexErrMsgTxt("Only dense real matrices are supported.");

		if (profile)
		{
			stringstream ss;
			printTimers(ss);
			mexPrintf("%s", ss.str().c_str());
		}

}