#include <iostream>
#include <fstream>
#include <limits>
#include "QUIC.h"
#include "Timer.h"

typedef float Float;

#define RANDOM_SEED 12345

#define TEST_GPU
//#define TEST_BLOCK_GPU
//#define TEST_MEM_BLOCK_GPU
//#define TEST_CPU
//#define TEST_BLOCK_CPU
//#define TEST_MEM_BLOCK_CPU
//#define TEST_LAPACK
#define ERROR_CHECK
#define EPSILON 0.00001f
#define DELTA 0.01f
#define ORTHO_MODE_CPU QUIC::BLOCK_MGS
#define ORTHO_MODE_GPU QUIC::BLOCK_MGS
#define N_BLOCKS_CPU 5
#define N_BLOCKS_GPU 5
#define MAX_RANK 120

//#ifndef RANDOM_SEED
#include <time.h>
//#endif

using namespace std;
using namespace QUIC;

template <typename Float>
void testOrthogonality(const QUIC::Matrix<Float>& m, bool transposed)
{
	QUIC::Uint size = transposed ? m.nRows : m.nCols;
	QUIC::Matrix<Float> mTm(size, size);
	QUIC::Matrix<Float>::mul(m, m, mTm, !transposed, transposed);
	Float diagMin = numeric_limits<Float>::infinity();
	Float offDiagMin = numeric_limits<Float>::infinity();
	Float diagMax = -numeric_limits<Float>::infinity();
	Float offDiagMax = -numeric_limits<Float>::infinity();
	double diagSum = 0.0;
	double diagSumSq = 0.0;
	double offDiagSum = 0.0;
	double offDiagSumSq = 0.0;
	int nDiag = 0;
	int nOffDiag = 0;
	for (int r=0; r<size; ++r)
	{
		Float * row = mTm.row(r);
		for (int c=0; c<size; ++c)
		{
			const Float entry = row[c];
			if (r == c)
			{
				diagSum += (double)entry;
				diagSumSq += (double)entry*entry;
				if (entry < diagMin) diagMin = entry;
				if (entry > diagMax) diagMax = entry;
				++nDiag;
			}
			else
			{
				offDiagSum += (double)entry;
				offDiagSumSq += (double)entry*entry;
				if (entry < offDiagMin) offDiagMin = entry;
				if (entry > offDiagMax) offDiagMax = entry;
				++nOffDiag;
			}
		}
	}
	double diagMean = diagSum/(double)nDiag;
	double diagVariance = diagSumSq/(double)nDiag - diagMean*diagMean;
	double offDiagMean = offDiagSum/(double)nOffDiag;
	double offDiagVariance = offDiagSumSq/(double)nOffDiag - offDiagMean*offDiagMean;
	cout<<"Diagonal:     Mean: "<<diagMean<<" Variance: "<<diagVariance<<" Min: "<<diagMin<<" Max: "<<diagMax<<endl
		<<"Off Diagonal: Mean: "<<offDiagMean<<" Variance: "<<offDiagVariance<<" Min: "<<offDiagMin<<" Max: "<<offDiagMax<<endl;
}

template <typename Float>
void checkError(const QUIC::Matrix<Float>& A, const QUIC::Matrix<Float>& U, const QUIC::Matrix<Float>& S, const QUIC::Matrix<Float>& VT)
{
	QUIC::Matrix<Float> Smat(S.nCols, S.nCols);
	for (int r=0; r<Smat.nRows; ++r) 
	{
		Float * row = Smat.row(r);
		for (int c=0; c<Smat.nCols; ++c)
		{
			row[c] = 0.0f;
		}
	}
	for (int i=0; i<S.nCols; ++i)
	{
		const Float * e1 = S.row(0) + i;
		Float * e2 = Smat.row(i) + i;
		*e2 = *e1;
	}
	QUIC::Matrix<Float> US(U.nRows, Smat.nCols);
	QUIC::Matrix<Float>::mul(U, Smat, US, false, false);
	QUIC::Matrix<Float> A2(A.nRows, A.nCols);
	QUIC::Matrix<Float>::mul(US, VT, A2, false, false);
	for (int r=0; r<A.nRows; ++r)
	{
		const Float * rowA = A.row(r);
		Float * rowA2 = A2.row(r);
		for (int c=0; c<A.nCols; ++c)
		{
			rowA2[c] -= rowA[c]; 
		}
	}
	double n1 = A2.norm();
	double n2 = A.norm();
	cout<<"Error: "<<(n1*n1)/(n2*n2)<<endl;
	cout<<"UTU:"<<endl;
	testOrthogonality(U, false);
	cout<<"VTV:"<<endl;
	testOrthogonality(VT, true);
}

template <typename T>
inline T square(const T& t)
{
	return t*t;
}

template <typename Float>
void checkErrorBlock(ifstream& ifs, QUIC::Matrix<Float>& U, QUIC::Matrix<Float>& S, QUIC::Matrix<Float>& VT, QUIC::Uint nBlocks)
{
	QUIC::Matrix<Float> * Smat = new QUIC::Matrix<Float>(S.nCols, S.nCols);
	for (int r=0; r<Smat->nRows; ++r) 
	{
		Float * row = Smat->row(r);
		for (int c=0; c<Smat->nCols; ++c)
		{
			row[c] = 0.0f;
		}
	}
	for (int i=0; i<S.nCols; ++i)
	{
		const Float * e1 = S.row(0) + i;
		Float * e2 = Smat->row(i) + i;
		*e2 = *e1;
	}
	QUIC::Matrix<Float> SVT(Smat->nRows, VT.nCols);
	QUIC::Matrix<Float>::mul(*Smat, VT, SVT, false, false);
	delete Smat;
	std::string format;
	getline(ifs, format, '\0');
	const size_t L = strlen("float");
	if (format.size() > 5 && format.substr(0, L) == "float")
	{
		size_t floatSize = static_cast<size_t>(atoi(format.c_str() + L));
		if (floatSize != 8uLL*sizeof(Float)) 
		{
			cout<<"Invalid data type"<<endl;
			return;
		}
		int size [2];
		ifs.read(reinterpret_cast<char *>(size), 2*sizeof(int));
		QUIC::Uint nRows = static_cast<QUIC::Uint>(size[0]);
		QUIC::Uint nCols = static_cast<QUIC::Uint>(size[1]);
		QUIC::Uint partitionNRows = (nRows/nBlocks) + (nRows%nBlocks ? 1 : 0);
		QUIC::Uint partitionSize = nCols*partitionNRows;
		QUIC::Uint partitionStart = 0uLL;
		Matrix<Float> partition(partitionNRows, nCols);
		Matrix<Float> svdSubset(partitionNRows, nCols);
		double Anorm = 0.0;
		double diffNorm = 0.0;
		for (QUIC::Uint i=0; i<nBlocks; ++i)
		{
			QUIC::Uint partitionSize = partitionNRows;
			QUIC::Uint partitionEnd = std::min(partitionStart + partitionNRows, nRows);
			partition.nRows = partitionEnd - partitionStart;
			partition.readFile<Float>(ifs);
			Matrix<Float> Usubset(partition.nRows, nCols, U.row(partitionStart), U.pitch);
			partitionStart += partitionNRows;
			Anorm += square(partition.norm());
			QUIC::Matrix<Float>::mul(Usubset, SVT, svdSubset, false, false);
			for (int r=0; r<partition.nRows; ++r)
			{
				const Float * rowA = partition.row(r);
				Float * rowSVD = svdSubset.row(r);
				for (int c=0; c<partition.nCols; ++c)
				{
					rowSVD[c] -= rowA[c]; 
				}
			}	
			diffNorm += square(svdSubset.norm());
		}
		cout<<"Error: "<<diffNorm/Anorm<<endl;
	}
}

void testBlockGPU(ifstream& ifs, Float epsilon, Float delta)
{
	cout<<"Block GPU:"<<endl;
	const streampos filePos = ifs.tellg();
	MatrixGPU<Float> gpuU, gpuS, gpuVT;
	if (QUIC::checkError())
	{
		cout<<"Error: "<<*getLastError()<<endl;
		return;
	}
	QUIC::Uint gpuN = QUIC_SVD<Float>(ifs, &gpuU, &gpuS, &gpuVT, epsilon, delta, ORTHO_MODE_GPU, N_BLOCKS_GPU, MAX_RANK);
	ifs.seekg(filePos);
	if (gpuN == 0) 
	{
		cout<<"Error: "<<*getLastError()<<endl;
		return;
	}
	else
	{
		cout<<gpuN<<" basis vectors."<<endl;
#ifdef ERROR_CHECK
		ifs.seekg(filePos);
		Matrix<Float> U(gpuU);
		Matrix<Float> S(gpuS);
		Matrix<Float> VT(gpuVT);
		checkErrorBlock(ifs, U, S, VT, N_BLOCKS_GPU);
		cout<<endl;
		ifs.seekg(filePos);
#endif
	}
}

void testBlockCPU(ifstream& ifs, Float epsilon, Float delta)
{
	cout<<"Block CPU:"<<endl;
	const streampos filePos = ifs.tellg();
	Matrix<Float> U, S, VT;
	QUIC::Uint cpuN = QUIC_SVD<Float>(ifs, &U, &S, &VT, epsilon, delta, ORTHO_MODE_CPU, N_BLOCKS_CPU, MAX_RANK);
	ifs.seekg(filePos);
	if (cpuN == 0) 
	{
		cout<<"Error: "<<*getLastError()<<endl;
		return;
	}
	else
	{
		cout<<cpuN<<" basis vectors."<<endl;
#ifdef ERROR_CHECK
		ifs.seekg(filePos);
		checkErrorBlock(ifs, U, S, VT, N_BLOCKS_CPU);
		cout<<endl;
		ifs.seekg(filePos);
#endif
	}
}

void testMemBlockGPU(const QUIC::Matrix<Float>& A, Float epsilon, Float delta)
{
	cout<<"Block GPU (Memory):"<<endl;
	MatrixGPU<Float> gpuU, gpuS, gpuVT;
	if (QUIC::checkError())
	{
		cout<<"Error: "<<*getLastError()<<endl;
		return;
	}
	clock_t start = clock();
	QUIC::Uint gpuN = QUIC_SVD<Float>(&A, &gpuU, &gpuS, &gpuVT, epsilon, delta, ORTHO_MODE_GPU, N_BLOCKS_GPU, MAX_RANK);
	clock_t end = clock();
	if (gpuN == 0) 
	{
		cout<<"Error: "<<*getLastError()<<endl;
		return;
	}
	else
	{
		cout<<gpuN<<" basis vectors."<<endl;
		cout<<(end - start)<<" ms."<<endl;
#ifdef ERROR_CHECK
		checkError(A, Matrix<Float>(gpuU), Matrix<Float>(gpuS), Matrix<Float>(gpuVT));
		cout<<endl;
#endif
	}
}

void testMemBlockCPU(const QUIC::Matrix<Float>& A, Float epsilon, Float delta)
{
	cout<<"Block CPU (Memory):"<<endl;
	Matrix<Float> U, S, VT;
	QUIC::Uint cpuN = QUIC_SVD<Float>(&A, &U, &S, &VT, epsilon, delta, ORTHO_MODE_CPU, N_BLOCKS_CPU, MAX_RANK);
	if (cpuN == 0) 
	{
		cout<<"Error: "<<*getLastError()<<endl;
		return;
	}
	else
	{
		cout<<cpuN<<" basis vectors."<<endl;
#ifdef ERROR_CHECK
		checkError(A, U, S, VT);
		cout<<endl;
#endif
	}
}

void testGPU(const QUIC::Matrix<Float>& A, Float epsilon, Float delta)
{
	cout<<"GPU:"<<endl;
	MatrixGPU<Float> Agpu(A);
	if (QUIC::checkError())
	{
		cout<<"Error: "<<*getLastError()<<endl;
		return;
	}
	MatrixGPU<Float> gpuU, gpuS, gpuVT;
	if (QUIC::checkError())
	{
		cout<<"Error: "<<*getLastError()<<endl;
		return;
	}
	QUIC::Uint gpuN = QUIC_SVD<Float>(&Agpu, &gpuU, &gpuS, &gpuVT, epsilon, delta, ORTHO_MODE_GPU, MAX_RANK);
	if (gpuN == 0) 
	{
		cout<<"Error: "<<*getLastError()<<endl;
		return;
	}
	else
	{
		cout<<gpuN<<" basis vectors."<<endl;
#ifdef ERROR_CHECK
		checkError(A, Matrix<Float>(gpuU), Matrix<Float>(gpuS), Matrix<Float>(gpuVT));
		cout<<endl;

		/* print out singular values */
		Matrix<Float> cpuS(gpuS);
		
		for (int i=0; i<cpuS.nCols; i++) {
			cout << cpuS.data[i] << endl;
		}
		
#endif
	}
}

void testCPU(const QUIC::Matrix<Float>& A, Float epsilon, Float delta)
{
	cout<<"CPU:"<<endl;
	Matrix<Float> U, S, VT;
	QUIC::Uint cpuN = QUIC_SVD<Float>(&A, &U, &S, &VT, epsilon, delta, ORTHO_MODE_CPU, MAX_RANK);
	if (cpuN == 0) 
	{
		cout<<"Error: "<<*getLastError()<<endl;
		return;
	}
	else
	{
		cout<<cpuN<<" basis vectors."<<endl;
#ifdef ERROR_CHECK
		checkError(A, U, S, VT);
		cout<<endl;
#endif
		/* print out singular values */
		printf("%16f, %16f, %16f, %16f\n", A.data[0], A.data[1], A.data[2], A.data[3]);
		for (int i=0; i<S.nCols; i++) {
			cout << S.data[i] << endl;
		}
	}
}

void testLAPACK(Matrix<Float>& A)
{
	Matrix<Float> UL, SL, VTL;
	Matrix<Float> temp(A);
	Timer T;
	T.start();
	temp.svd(UL, SL, VTL);
	T.stop();
	cout<<(static_cast<Float>(T.ms())/1000.0f)<<" seconds"<<endl;
#ifdef ERROR_CHECK
	checkError(A, UL, SL, VTL);
	cout<<endl;
#endif
}

int main(int argc, char * argv [])
{
	//ifstream inputFile("input.txt");
	//if (!inputFile.is_open()) exit(1);
	//string filename;
	//getline(inputFile, filename);
	//inputFile.close();
	if (argc < 2) {
		cout<<"Usage: "<<argv[0]<<" input.bin"<<endl;
		exit(1);
	}
	string filename = argv[1];
	ifstream ifs(filename.c_str(), ios::binary);
	if (!ifs.is_open())
	{
		cout<<"Failed to open "<<filename<<endl;
		exit(1);
	}
#ifdef RANDOM_SEED
	const unsigned int randomSeed = RANDOM_SEED;
#else
	const unsigned int randomSeed = (unsigned int)time(NULL);
#endif
	cout<<"Random seed: "<<randomSeed<<endl;
	Float epsilon = EPSILON;
	Float delta = DELTA;

#ifdef TEST_BLOCK_GPU
	srand(randomSeed);
	initGPU();
	testBlockGPU(ifs, epsilon, delta);
	shutdownGPU();
	if (checkError())
	{
		cout<<*getLastError()<<endl;
#ifdef _DEBUG
		cin.get();
#endif
		exit(1);
	}
#endif

#ifdef TEST_BLOCK_CPU
	srand(randomSeed);
	testBlockCPU(ifs, epsilon, delta);
	if (checkError())
	{
		cout<<*getLastError()<<endl;
#ifdef _DEBUG
		cin.get();
#endif
		exit(1);
	}
#endif
#if defined TEST_CPU || defined TEST_GPU || defined TEST_MEM_BLOCK_CPU || defined TEST_MEM_BLOCK_GPU
	Matrix<Float> A(ifs);
	ifs.close();
	if (checkError())
	{
		cout<<*getLastError()<<endl;
#ifdef _DEBUG
		cin.get();
#endif
		exit(1);
	}

#ifdef TEST_MEM_BLOCK_GPU
	srand(randomSeed);
	initGPU();
	testMemBlockGPU(A, epsilon, delta);
	shutdownGPU();
	if (checkError())
	{
		cout<<*getLastError()<<endl;
#ifdef _DEBUG
		cin.get();
#endif
		exit(1);
	}
#endif

#ifdef TEST_MEM_BLOCK_CPU
	srand(randomSeed);
	testMemBlockCPU(A, epsilon, delta);
	if (checkError())
	{
		cout<<*getLastError()<<endl;
#ifdef _DEBUG
		cin.get();
#endif
		exit(1);
	}
#endif

#ifdef TEST_GPU
	cout<<"Matrix size: "<<A.nRows<<" x "<<A.nCols<<endl;
	srand(randomSeed);
	initGPU();
	testGPU(A, epsilon, delta);
	shutdownGPU();
	if (checkError())
	{
		cout<<*getLastError()<<endl;
#ifdef _DEBUG
		cin.get();
#endif
		exit(1);
	}
#endif

#ifdef TEST_CPU
	srand(randomSeed);
	testCPU(A, epsilon, delta);
	if (checkError())
	{
		cout<<*getLastError()<<endl;
#ifdef _DEBUG
		cin.get();
#endif
		exit(1);
	}
#endif

#ifdef TEST_LAPACK
	testLAPACK(A);
#endif
	printTimers(cout);
#ifdef _DEBUG
	cin.get();
#endif
#endif
}